#############################################################################
# Author: <kerstin.weger@xfel.eu>
# Created on November 3, 2011
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from qtpy.QtCore import Qt, Slot
from qtpy.QtGui import QCursor, QPalette
from qtpy.QtWidgets import (
    QAction, QDialog, QHBoxLayout, QPushButton, QScrollArea, QStackedWidget,
    QToolButton, QTreeView, QVBoxLayout, QWidget)

import karabogui.access as krb_access
from karabo.native import AccessMode, Assignment, Hash, has_changes
from karabogui import icons, messagebox
from karabogui.access import AccessRole, access_role_allowed
from karabogui.binding.api import (
    DeviceClassProxy, DeviceProxy, ProjectDeviceProxy, VectorHashBinding,
    apply_configuration, get_binding_value, validate_table_value,
    validate_value)
from karabogui.configurator.api import ConfigurationTreeView
from karabogui.dialogs.api import (
    ConfigurationFromPastDialog, ConfigurationFromPastPreview,
    DeviceSelectorDialog, InitConfigurationPreviewDialog)
from karabogui.events import (
    KaraboEvent, register_for_broadcasts, unregister_from_broadcasts)
from karabogui.logger import get_logger
from karabogui.singletons.api import get_manager
from karabogui.util import (
    get_spin_widget, load_configuration_from_file, move_to_cursor,
    save_configuration_to_file)
from karabogui.widgets.toolbar import ToolBar

from .base import BasePanelWidget
from .panel_info import create_configurator_info
from .switch import SwitchButton
from .tool_widget import ConfiguratorSearch
from .utils import (
    compare_proxy_essential, format_property_details,
    format_vector_hash_details)

BLANK_PAGE = 0
WAITING_PAGE = 1
CONFIGURATION_PAGE = 2


class ConfigurationPanel(BasePanelWidget):
    def __init__(self, allow_closing=False):
        super().__init__("Configuration Editor", allow_closing)
        self._showing_proxy = None
        self._awaiting_schema = None

        # Register for broadcast events.
        self.event_map = {
            KaraboEvent.ShowConfiguration: self._event_show_configuration,
            KaraboEvent.UpdateDeviceConfigurator: self._event_update_config,
            KaraboEvent.UpdateValueConfigurator: self._event_display_update,
            KaraboEvent.ClearConfigurator: self._event_clear_configurator,
            KaraboEvent.LoadConfiguration: self._event_load_configuration,
            KaraboEvent.ShowConfigurationFromPast: self._event_config_past,
            KaraboEvent.ShowInitConfiguration: self._event_init_config,
            KaraboEvent.NetworkConnectStatus: self._event_network,
            KaraboEvent.AccessLevelChanged: self._event_access_level,
            KaraboEvent.LoginUserChanged: self._event_access_level,
        }
        register_for_broadcasts(self.event_map)

    # -----------------------------------------------------------------------
    # Karabo Events

    def _event_access_level(self, data):
        self._update_buttons(self._showing_proxy)
        access = krb_access.GLOBAL_ACCESS_LEVEL
        self.treeView().setAccessLevel(access)

    def _event_show_configuration(self, data):
        proxy = data['proxy']
        self._show_configuration(proxy)

    def _event_update_config(self, data):
        proxy = data['proxy']
        refresh = data.get('refresh', False)
        self._update_displayed_configuration(proxy, refresh)

    def _event_display_update(self, data):
        proxy = data['proxy']
        self._update_displayed_values(proxy)

    def _event_clear_configurator(self, data):
        devices = data.get('devices', [])
        self._remove_departed_device(devices)

    def _event_load_configuration(self, data):
        proxy = data['proxy']
        configuration = data['configuration']
        self._apply_loaded_configuration(proxy, configuration)

    def _event_config_past(self, data):
        deviceId = data['deviceId']
        configuration = data['configuration']
        req_time = data['time']
        config_time = data['config_time']
        time_match = data['time_match']
        preview = data['preview']
        self._apply_configuration_from_past(deviceId, configuration, req_time,
                                            config_time, time_match, preview)

    def _event_init_config(self, data):
        deviceId = data['deviceId']
        configuration = data['configuration']
        name = data['name']
        preview = data['preview']
        classId = data['classId']
        self._apply_init_configuration(
            deviceId, classId, configuration, name, preview)

    def _event_network(self, data):
        connected = data['status']
        if not connected:
            self._reset_panel()

    # -----------------------------------------------------------------------

    def get_content_widget(self):
        """Returns a QWidget containing the main content of the panel.
        """
        widget = QWidget(self)
        mainLayout = QVBoxLayout(widget)
        mainLayout.setContentsMargins(0, 0, 0, 2)

        # Stacked widget for configuration parameters
        self._stacked_tree_widgets = QStackedWidget(widget)
        # BLANK_PAGE
        self._stacked_tree_widgets.addWidget(QTreeView(widget))

        # WAITING_PAGE
        wait_widget = get_spin_widget(icon='wait', parent=widget)
        wait_widget.setAlignment(Qt.AlignHCenter | Qt.AlignVCenter)
        wait_widget.setAutoFillBackground(True)
        wait_widget.setBackgroundRole(QPalette.Base)
        self._stacked_tree_widgets.addWidget(wait_widget)

        # CONFIGURATION_PAGE
        editor = ConfigurationTreeView(widget)
        editor.sourceModel().signalHasModifications.connect(
            self._on_button_changes)
        self._stacked_tree_widgets.addWidget(editor)

        self.searchBar = ConfiguratorSearch()
        self.searchBar.setView(editor)
        mainLayout.addWidget(self.searchBar)

        hLayout = QHBoxLayout()
        hLayout.setContentsMargins(0, 5, 5, 5)

        text = "Instantiate device"
        self.pbInitDevice = QPushButton(icons.run, text)
        self.pbInitDevice.setToolTip(text)
        self.pbInitDevice.setStatusTip(text)
        self.pbInitDevice.setVisible(False)
        self.pbInitDevice.setMinimumWidth(140)
        self.pbInitDevice.clicked.connect(self._on_init_device)
        hLayout.addWidget(self.pbInitDevice)

        text = "Shutdown instance"
        self.pbKillInstance = QPushButton(icons.kill, text)
        self.pbKillInstance.setStatusTip(text)
        self.pbKillInstance.setToolTip(text)
        self.pbKillInstance.setVisible(False)
        self.pbKillInstance.setMinimumWidth(140)
        # use action for button to reuse
        self.acKillInstance = QAction(icons.kill, text, widget)
        self.acKillInstance.setStatusTip(text)
        self.acKillInstance.setToolTip(text)
        self.acKillInstance.triggered.connect(self._on_kill_device)
        self.pbKillInstance.clicked.connect(self.acKillInstance.triggered)
        self.pbKillInstance.setEnabled(not self.allow_closing)
        hLayout.addWidget(self.pbKillInstance)

        text = "Apply all"
        description = "Apply all property changes in one go"
        self.pbApplyAll = QPushButton(icons.apply, text)
        self.pbApplyAll.setToolTip(description)
        self.pbApplyAll.setStatusTip(description)
        self.pbApplyAll.setVisible(False)
        self.pbApplyAll.setEnabled(False)
        self.pbApplyAll.setMinimumWidth(140)
        # use action for button to reuse
        self.acApplyAll = QAction(icons.apply, text, widget)
        self.acApplyAll.setStatusTip(text)
        self.acApplyAll.setToolTip(text)
        self.acApplyAll.triggered.connect(self._on_apply_all)
        self.pbApplyAll.clicked.connect(self.acApplyAll.triggered)
        hLayout.addWidget(self.pbApplyAll)

        text = "Decline all"
        decription = ("Decline all property changes and reset them to value "
                      "on device")
        self.pbResetAll = QPushButton(icons.no, text)
        self.pbResetAll.setToolTip(decription)
        self.pbResetAll.setStatusTip(decription)
        self.pbResetAll.setVisible(False)
        self.pbResetAll.setEnabled(False)
        self.pbResetAll.setMinimumWidth(140)
        # use action for button to reuse
        self.acResetAll = QAction(icons.no, text, widget)
        self.acResetAll.setStatusTip(text)
        self.acResetAll.setToolTip(text)
        self.acResetAll.triggered.connect(self._on_reset_all)
        self.pbResetAll.clicked.connect(self.acResetAll.triggered)
        hLayout.addWidget(self.pbResetAll)
        hLayout.addStretch()

        self.rightScrollArea = QScrollArea(widget)
        self.rightScrollArea.setWidgetResizable(True)
        rightWidget = QWidget(widget)
        vLayout = QVBoxLayout(rightWidget)
        vLayout.setContentsMargins(0, 0, 0, 0)
        vLayout.addWidget(self._stacked_tree_widgets)
        vLayout.addLayout(hLayout)
        self.rightScrollArea.setWidget(rightWidget)
        mainLayout.addWidget(self.rightScrollArea)

        return widget

    def toolbars(self):
        """This should create and return one or more `ToolBar` instances needed
        by this panel.
        """
        toolbar = ToolBar(parent=self)
        icon_size = toolbar.iconSize()

        text = "Open configuration from file (*.xml)"
        tb_open_config = QPushButton()
        tb_open_config.setIcon(icons.load)
        tb_open_config.setIconSize(icon_size)
        tb_open_config.setStatusTip(text)
        tb_open_config.setToolTip(text)
        tb_open_config.setVisible(False)
        tb_open_config.setFlat(True)
        tb_open_config.clicked.connect(self._on_open_from_file)

        text = "Save configuration to file (*.xml)"
        tb_save_config = QPushButton()
        tb_save_config.setIcon(icons.saveAs)
        tb_save_config.setIconSize(icon_size)
        tb_save_config.setStatusTip(text)
        tb_save_config.setToolTip(text)
        tb_save_config.setVisible(False)
        tb_save_config.setFlat(True)
        tb_save_config.clicked.connect(self._on_save_to_file)

        text = "Runtime configuration"
        tb_runtime_config = QToolButton()
        tb_runtime_config.setIcon(icons.clock)
        tb_runtime_config.setStatusTip(text)
        tb_runtime_config.setToolTip(text)
        tb_runtime_config.setVisible(False)
        tb_runtime_config.clicked.connect(self._on_config_from_past)

        text = "Compare Device"
        tb_compare_device = QToolButton()
        tb_compare_device.setIcon(icons.change)
        tb_compare_device.setStatusTip(text)
        tb_compare_device.setToolTip(text)
        tb_compare_device.setVisible(False)
        tb_compare_device.clicked.connect(self._on_compare_configuration)

        text = "Switch Configuration"
        tb_switch_mode = SwitchButton()
        tb_switch_mode.setStatusTip(text)
        tb_switch_mode.setToolTip(text)
        tb_switch_mode.setVisible(True)
        tb_switch_mode.setMinimumWidth(140)
        tb_switch_mode.toggled.connect(self._switch_mode_toggled)

        text = "Search Device Properties"
        tb_search_device = QToolButton()
        tb_search_device.setIcon(icons.zoomImage)
        tb_search_device.setIconSize(icon_size)
        tb_search_device.setStatusTip(text)
        tb_search_device.setToolTip(text)
        tb_search_device.setVisible(True)
        tb_search_device.clicked.connect(self.searchBar.toggleActivate)

        self.ui_open_config = toolbar.addWidget(tb_open_config)
        self.ui_save_config = toolbar.addWidget(tb_save_config)
        self.ui_history_config = toolbar.addWidget(tb_runtime_config)
        self.ui_compare_config = toolbar.addWidget(tb_compare_device)
        self.ui_search_device = toolbar.addWidget(tb_search_device)
        self.ui_switch_mode = toolbar.addWidget(tb_switch_mode)

        return [toolbar]

    # -----------------------------------------------------------------------
    # private methods

    def _check_configuration(self, deviceId, classId):
        proxy = self._showing_proxy
        if proxy is None:
            return False
        if not isinstance(proxy, (DeviceProxy, ProjectDeviceProxy)):
            msg = (f"A configuration for '{deviceId}' arrived, but is "
                   "ignored since the device is not shown in editor.")
            messagebox.show_error(msg, parent=self)
            return False

        binding = proxy.binding
        # The check we can provide is to check the deviceId and classId
        if classId is None:
            # XXX: We might still get an invalid configuration without classId
            messagebox.show_error("A configuration without classId arrived "
                                  "and is ignored! ", parent=self)
            return False

        if proxy.device_id != deviceId:
            messagebox.show_error(
                "A configuration for '{}' arrived, but is "
                "ignored since '{}' shown in editor.".format(
                    deviceId, proxy.device_id), parent=self)
            return False
        if binding.class_id != classId:
            # Note: Previously we validated strong here. Since we validate
            # binding by binding, we continue gracefully but leave a log
            # message. Schema evolution is not a problem anymore.
            text = (f"A configuration for classId <b>{classId}</b> arrived, "
                    f"but the device <b>{deviceId}</b> shown in editor is a "
                    f"<b>{binding.class_id}</b>")
            get_logger().error(text)
            return True

        return True

    def _apply_init_configuration(
            self, deviceId, classId, config, name, preview):
        """Apply the retrieved configuration from getConfigurationFromName"""
        validated = self._check_configuration(deviceId, classId)
        if not validated:
            return

        proxy = self._showing_proxy
        if preview:
            title = (f"Preview init configuration '{name}' for device "
                     f"'{proxy.device_id}'")
            dialog = InitConfigurationPreviewDialog(
                self._showing_proxy, config, title, parent=self)
            if dialog.exec() != QDialog.Accepted:
                return
            config = dialog.configuration_to_apply
        else:
            text = (f"Configuration with name <b>{name}</b> has arrived for <b"
                    f">{deviceId}</b>.\n")
            messagebox.show_information(text, parent=self)

        self._set_proxy_configuration(proxy, config)

    def _apply_configuration_from_past(self, deviceId, config, req_time,
                                       config_time, match, preview):
        """Apply the retrieved configuration from getConfigurationFromPast"""
        classId = config.get("classId")
        validated = self._check_configuration(deviceId, classId)
        if not validated:
            return

        proxy = self._showing_proxy
        # Show a nice information for the user!
        matched_text = ("The configuration has been retrieved for the "
                        "requested time point!")
        no_match_text = ("Requested time was '{}', but the device was not "
                         "logged then.").format(req_time)
        time_text = matched_text if match else no_match_text
        text = "Configuration from '{}' has arrived for '{}'.\n{}".format(
            config_time, deviceId, time_text)

        if preview:
            title = f"Showing configuration for device {proxy.device_id}"
            dialog = ConfigurationFromPastPreview(title, text, config, proxy,
                                                  parent=self)
            if dialog.exec() == QDialog.Accepted:
                self._set_proxy_configuration(proxy, config)
        else:
            self._set_proxy_configuration(proxy, config)
            messagebox.show_information(text, parent=self)

    def _apply_loaded_configuration(self, proxy, configuration):
        """Apply a configuration loaded from a file to a proxy
        """
        if self._showing_proxy is not proxy:
            return

        binding = proxy.binding
        # Loading a configuration from file has to check for a present Schema
        # and if the classId is the first key of the configuration
        if len(binding.value) == 0 or binding.class_id not in configuration:
            messagebox.show_error('Configuration load failed in configurator.',
                                  parent=self)
            return

        configuration = configuration[binding.class_id]
        self._set_proxy_configuration(proxy, configuration)

    def _set_proxy_configuration(self, proxy, configuration):
        """Internal method to apply the configuration in the Configurator
        """
        if isinstance(proxy, DeviceProxy):
            self._set_device_config(proxy, configuration)
        else:
            self._set_project_device_config(proxy, configuration)

    def _set_device_config(self, proxy, configuration):
        """Sets the configuration of an online device by validating and setting
           the edit value of every reconfigurable property. This is done as
           follows:

           1. Iterate over each property
           2. Check if the property binding is valid
           3. Check if the binding can be edited
              (reconfigurable, allowed state and allowed access level)
           4. Check if there are changes in the value (first check)
           5. If there are changes, we validate the value from the binding
              information. If it fails to be validated, we report it as invalid
           6. We again compare to check changes (second check)
           7. If there are changes, we set it on the edit value.
        """
        # Load the configuration into PropertyProxy instances
        invalid_prop = {}
        access_level = krb_access.GLOBAL_ACCESS_LEVEL
        state = proxy.state_binding.value

        editor = self._stacked_tree_widgets.widget(CONFIGURATION_PAGE)
        model = editor.sourceModel()
        for path, value, _ in Hash.flat_iterall(configuration):
            prop_proxy = model.property_proxy(path)
            prop_binding = prop_proxy.binding
            if prop_binding is None:
                # NOTE: This property most likely was removed from the
                # device, we have schema evolution and will continue here!
                continue
            if prop_binding.accessMode is not AccessMode.RECONFIGURABLE:
                continue
            prop_value = get_binding_value(prop_binding)
            edit_value = None
            if (prop_binding.requiredAccessLevel <= access_level
                    and prop_binding.is_allowed(state)):
                if isinstance(prop_binding, VectorHashBinding):
                    valid, invalid = validate_table_value(prop_binding, value)
                    if (valid and has_changes(
                            prop_value, valid)):
                        edit_value = valid
                    if invalid:
                        invalid_prop[path] = invalid
                elif has_changes(prop_value, value):
                    valid = validate_value(prop_binding, value)
                    if valid is None:
                        invalid_prop[path] = value
                    elif has_changes(prop_value, valid):
                        edit_value = valid
            prop_proxy.edit_value = edit_value
        # Note: We tell the model directly to send dataChanged signal and
        # notify for changes!
        model.announceDataChanged()
        model.notify_of_modifications()

        # Show a dialog for invalid keys:
        if invalid_prop:
            self._show_not_loaded_properties(proxy, invalid_prop)

    def _set_project_device_config(self, proxy, configuration):
        """Sets the configuration of an of device by validating and setting
           the value of every non-readonly property. This is done as follows:

           1. Iterate over each writable property by (deep)copying the
              configuration and deleting nonexistent and non-configurable
              properties to exclude it from bulk setting
           2. Check if the property binding is valid and writeable
              (non-readonly). If not, we exclude the property from the bulk
              setting by deleting it from the configuration
           3. Validate the value from the binding information. If it fails to
              be validated, we report it as invalid and exclude it from
              the bulk setting.
           4. Apply the validated configuration and its attributes
        """

        # Get the validated config for the bulk update
        valid, invalid = self._get_validated_config(proxy, configuration)

        # Load the configuration directly into the binding
        if len(valid):
            apply_configuration(valid, proxy.binding)
            # Notify again
            proxy.binding.config_update = True

        # Show a dialog for invalid keys:
        if invalid:
            self._show_not_loaded_properties(proxy, invalid)

    def _get_validated_config(self, proxy, configuration):
        """Validate the configuration by populating a new Hash with the
           validated values and attributes. We ignore read-only properties
           and with invalid values"""

        valid, invalid = Hash(), Hash()  # {path: value}
        for path, value, attrs in Hash.flat_iterall(configuration):
            binding = proxy.get_property_binding(path)
            writable = (binding is not None and
                        binding.accessMode != AccessMode.READONLY and
                        binding.assignment != Assignment.INTERNAL)
            if not writable:
                continue
            if isinstance(binding, VectorHashBinding):
                valid_vhash, invalid_vhash = validate_table_value(
                    binding, value)
                if invalid_vhash:
                    invalid[path] = invalid_vhash
                if valid_vhash:
                    old_value = get_binding_value(binding)
                    if has_changes(old_value, valid_vhash):
                        valid[path] = valid_vhash
            else:
                validated_value = validate_value(binding, value)
                if validated_value is not None:
                    valid[path] = validated_value
                    valid[path, ...] = attrs
                else:
                    # Report only existing and configurable properties but
                    # with invalid values.
                    invalid[path] = value

        return valid, invalid

    def _show_not_loaded_properties(self, device_proxy, invalid_prop):
        msg = ("Some values are not loaded as they are "
               "invalid for the current property.")
        details = []
        for path, value in invalid_prop.items():
            binding = device_proxy.get_property_binding(path)
            if isinstance(binding, VectorHashBinding):
                value = format_vector_hash_details(binding, value)
            details.append(format_property_details(binding, path, value))

        messagebox.show_warning(msg, details="\n".join(details),
                                parent=self)

    def _update_buttons(self, proxy):
        """Main method to update the button visibility of the configurator"""
        access = access_role_allowed(AccessRole.SERVICE_EDIT)
        if (proxy is None or not len(proxy.binding.value)
                or (isinstance(proxy, DeviceClassProxy)
                    and not isinstance(proxy, ProjectDeviceProxy)) or
                not access):
            self._hide_all_buttons()
        else:
            is_device = isinstance(proxy, DeviceProxy)
            self._show_button_visibility(not (proxy is None or is_device))

        allow_fetch_config = isinstance(proxy,
                                        (DeviceProxy, ProjectDeviceProxy))
        self.ui_history_config.setVisible(allow_fetch_config)
        self.ui_compare_config.setVisible(allow_fetch_config and proxy.online)

    def _show_button_visibility(self, visible):
        """Show the configurator buttons depending of the `visible` parameter
        """
        self.pbInitDevice.setVisible(visible)

        self.pbKillInstance.setVisible(not visible)
        self.pbApplyAll.setVisible(not visible)
        self.pbResetAll.setVisible(not visible)

        self.acKillInstance.setVisible(not visible)
        self.acApplyAll.setVisible(not visible)
        self.acResetAll.setVisible(not visible)

    def _hide_all_buttons(self):
        """Hide buttons and actions"""
        self.pbInitDevice.setVisible(False)
        self.pbKillInstance.setVisible(False)
        self.acKillInstance.setVisible(False)
        self.pbApplyAll.setVisible(False)
        self.acApplyAll.setVisible(False)
        self.pbResetAll.setVisible(False)
        self.acResetAll.setVisible(False)

    def _remove_departed_device(self, devices):
        """Clear the configuration panel when a device goes away

        :param devices: List of devices that are gone!
        """
        proxy = self._showing_proxy
        if proxy is None or not isinstance(proxy, DeviceProxy):
            return

        if proxy.device_id in devices:
            self._show_configuration(None)

    def _reset_panel(self):
        """This is called when the configurator needs a reset which means all
        parameter editor pages need to be cleaned and removed.
        """
        self._showing_proxy = None

        tree_widget = self._stacked_tree_widgets.widget(CONFIGURATION_PAGE)
        tree_widget.clear()

        self._set_stack_widget_index(BLANK_PAGE)
        self._hide_all_buttons()

    def _schema_update(self, proxy, name, value):
        """Trait notification handler for a schema_update event on the
        BaseDeviceProxy object referenced in `self._awaiting_schema`.
        """
        assert proxy is self._awaiting_schema
        self._awaiting_schema = None
        proxy.on_trait_change(self._schema_update, 'schema_update',
                              remove=True)

        if self._showing_proxy is proxy:
            self._show_configuration(proxy)

    def _set_stack_widget_index(self, index):
        """Pick between one of the pages in `self._stacked_tree_widgets`.
        """
        self._stacked_tree_widgets.blockSignals(True)
        self._stacked_tree_widgets.setCurrentIndex(index)
        self._stacked_tree_widgets.blockSignals(False)

        visible = (index != BLANK_PAGE)
        self.ui_open_config.setVisible(visible)
        self.ui_save_config.setVisible(visible)
        self.ui_history_config.setVisible(visible)

    def _set_tree_widget_configuration(self, tree_widget, proxy):
        tree_widget.assign_proxy(proxy)
        tree_widget.resizeColumnToContents(0)
        tree_widget.resizeColumnToContents(1)

    def _set_configuration(self, proxy):
        """Adapt to the Configuration which is currently showing
        """
        # Update buttons at the bottom
        self._update_buttons(proxy)

        # Toggle device updates
        if (self._showing_proxy not in (None, proxy) and
                isinstance(self._showing_proxy, DeviceProxy)):
            self._showing_proxy.remove_monitor()

        if (proxy not in (None, self._showing_proxy) and
                isinstance(proxy, DeviceProxy)):
            proxy.add_monitor()

        # Cancel notification of arriving schema
        if self._awaiting_schema is not None:
            awaited = self._awaiting_schema
            awaited.on_trait_change(self._schema_update, 'schema_update',
                                    remove=True)
            self._awaiting_schema = None

        # Handle configurations which await a schema
        if proxy is not None and len(proxy.binding.value) == 0:
            self._awaiting_schema = proxy
            proxy.on_trait_change(self._schema_update, 'schema_update')

        # Finally, set _showing_proxy
        self._showing_proxy = proxy

    def _show_configuration(self, proxy):
        """Show a Configuration object in the panel
        """
        if proxy is None:
            # There is no configuration to show
            index = BLANK_PAGE
        elif len(proxy.binding.value) == 0:
            # The configuration is not ready to be shown (no Schema)
            index = WAITING_PAGE
        else:
            # The configuration is OK to show
            index = CONFIGURATION_PAGE
            tree_widget = self._stacked_tree_widgets.widget(index)
            self._set_tree_widget_configuration(tree_widget, proxy)

        # This is the configuration we're viewing now
        self._set_stack_widget_index(index)
        self._set_configuration(proxy)

    def _update_displayed_values(self, proxy):
        """Update the apply and decline buttons after displayed value update
        """
        if self._showing_proxy is not proxy:
            return

        index = CONFIGURATION_PAGE
        tree_widget = self._stacked_tree_widgets.widget(index)
        model = tree_widget.sourceModel()
        if model is not None:
            model.notify_of_modifications()

    def _update_displayed_configuration(self, proxy, refresh=False):
        """Update the displayed configuration for a `proxy`

        This can be enforced with `refresh` equal to `True`.
        """
        if self._showing_proxy is None:
            return
        if compare_proxy_essential(proxy, self._showing_proxy) or refresh:
            # - If current showing device proxy matches the updated one,
            #   refresh the view.
            # - The device might have been gone from online to offline. In case
            #   of a project device proxy, make sure we have a schema!
            # - We might have to refresh in case a device is renamed
            if isinstance(proxy, ProjectDeviceProxy):
                proxy.ensure_class_schema()
            self._show_configuration(proxy)

    # -----------------------------------------------------------------------
    # Public Interface

    def model(self):
        """Returns the `QAbstractItemModel` associated with the tree view"""
        tree_widget = self._stacked_tree_widgets.widget(CONFIGURATION_PAGE)
        return tree_widget.model()

    def treeView(self):
        """Returns the `QTreeView` associated with the configuration panel"""
        return self._stacked_tree_widgets.widget(CONFIGURATION_PAGE)

    def info(self):
        """Returns the panel data for the `KaraboLogBook`"""
        return create_configurator_info(self)

    def __repr__(self):
        return f"<{self.__class__.__name__} proxy={repr(self.model().root)}>"

    # -----------------------------------------------------------------------
    # Qt Slot

    @Slot(bool)
    def _on_button_changes(self, has_changes):
        self.pbApplyAll.setEnabled(has_changes)
        self.pbResetAll.setEnabled(has_changes)
        self.acApplyAll.setEnabled(has_changes)
        self.acResetAll.setEnabled(has_changes)

    @Slot()
    def _on_apply_all(self):
        tree_widget = self._stacked_tree_widgets.widget(CONFIGURATION_PAGE)
        tree_widget.apply_all()

    @Slot()
    def _on_reset_all(self):
        tree_widget = self._stacked_tree_widgets.widget(CONFIGURATION_PAGE)
        tree_widget.decline_all()

    @Slot()
    def _on_open_from_file(self):
        if self._showing_proxy is not None:
            load_configuration_from_file(self._showing_proxy, parent=self)

    @Slot()
    def _on_save_to_file(self):
        if self._showing_proxy is not None:
            save_configuration_to_file(self._showing_proxy, parent=self)

    @Slot()
    def _on_kill_device(self):
        if not isinstance(self._showing_proxy, DeviceProxy):
            return
        get_manager().shutdownDevice(self._showing_proxy.device_id,
                                     parent=self)

    @Slot()
    def _on_init_device(self):
        """Instantiate a project device from the configurator"""
        proxy = self._showing_proxy
        if not isinstance(proxy, ProjectDeviceProxy):
            # This should not happen, however we protect here.
            return

        serverId = proxy.server_id
        classId = proxy.binding.class_id
        deviceId = proxy.device_id
        get_manager().initDevice(serverId, classId, deviceId)

    @Slot()
    def _on_config_from_past(self) -> None:
        if not isinstance(self._showing_proxy,
                          (DeviceProxy, ProjectDeviceProxy)):
            return
        device_id = self._showing_proxy.device_id
        dialog = ConfigurationFromPastDialog(
            instance_id=device_id, parent=self)
        dialog.move(QCursor.pos())
        dialog.show()
        dialog.raise_()
        dialog.activateWindow()

    @Slot(bool)
    def _switch_mode_toggled(self, expert):
        """Change the mode setting of the tree widget"""
        tree_widget = self._stacked_tree_widgets.widget(CONFIGURATION_PAGE)
        tree_widget.setMode(expert)

    @Slot()
    def _on_compare_configuration(self) -> None:
        if not isinstance(self._showing_proxy,
                          (DeviceProxy, ProjectDeviceProxy)):
            return
        device_id = self._showing_proxy.device_id
        dialog = DeviceSelectorDialog(device_id, parent=self)
        move_to_cursor(dialog)
        dialog.show()

    # -----------------------------------------------------------------------
    # Qt

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            self.signalPanelClosed.emit(self.windowTitle())
            unregister_from_broadcasts(self.event_map)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from traits.api import pop_exception_handler, push_exception_handler

from karabo.common.project.api import MacroModel, ProjectModel
from karabogui.binding.api import ProxyStatus
from karabogui.singletons.mediator import Mediator
from karabogui.singletons.project_model import ProjectViewItemModel
from karabogui.testing import singletons

PATH_RUN_MACRO = "karabogui.project.controller.macro.run_macro"
MBOX_PATH = "karabogui.project.controller.macro.messagebox.show_warning"


def _get_subgroup_controllers(proj_controller, name):
    return [child for child in proj_controller.children
            if child.trait_name == name]


def _get_object_controllers(subgroup_controllers):
    controllers = []
    for subgroup in subgroup_controllers:
        controllers.extend(subgroup.children)
    return controllers


def test_macro_controller(gui_app, mocker):
    """Test the macro controller of the project"""
    push_exception_handler(lambda *args: None, reraise_exceptions=True)
    qt_model = ProjectViewItemModel(parent=None)

    # Instantiate a macro model
    simple_name = "macro_foo"
    uuid = "12345678-1234-5678-1234-567812345678"
    macro_model = MacroModel(simple_name=simple_name, uuid=uuid)
    macro_model.initialized = macro_model.modified = True

    def _assert_macro(code, is_valid):
        models = {"macros": [macro_model]}
        proj_model = ProjectModel(**models)

        # Cause the controllers to be created and get a ref to the root
        qt_model.root_model = proj_model
        proj_controller = qt_model.root_controller
        subgroup_controllers = _get_subgroup_controllers(proj_controller,
                                                         "macros")

        macro_controllers = _get_object_controllers(subgroup_controllers)
        macro_controller = macro_controllers[0]
        macro_model.code = code

        mocked_run_macro = mocker.patch(PATH_RUN_MACRO)
        mocked_messagebox = mocker.patch(MBOX_PATH)
        macro_controller.run_macro()
        if is_valid:
            mocked_run_macro.assert_called_once()
            mocked_messagebox.assert_not_called()
        else:
            mocked_run_macro.assert_not_called()
            mocked_messagebox.assert_called_once()

        class_id = simple_name.title()
        device_id = macro_model.instance_id
        devices = [(device_id, class_id, ProxyStatus.OFFLINE)]
        servers = []
        macro_controller.system_topology_callback(devices, servers)
        for child in macro_controller.children:
            assert child.ui_data.status is ProxyStatus.OFFLINE
            assert child.model is not macro_controller.model

    # Put here the assertion of the macro code
    with singletons(mediator=Mediator()):
        _assert_macro(VALID_MACRO, is_valid=True)
        _assert_macro(SYNTAX_ERROR_MACRO, is_valid=False)
        _assert_macro(IMPORT_ERROR_MACRO, is_valid=False)
        _assert_macro(INDENTATION_ERROR_MACRO, is_valid=False)

    pop_exception_handler()
    qt_model = None


VALID_MACRO = """
from karabo.middlelayer import Macro, Slot, String

class ValidMacro(Macro):
    name = String(defaultValue="Foo")

    @Slot()
    def execute(self):
        print("Hello {{}}!".format(name))

    @MacroSlot()
    def executeSlow(self):
        print("Hello {{}}!".format(name))
"""

SYNTAX_ERROR_MACRO = """
from karabo.middlelayer import Macro, Slot, String

class InvalidMacro(Macro)
    name = String(defaultValue="Foo")

    @Slot()
    def execute(self):
        print("Hello {{}}!".format(name))
"""

IMPORT_ERROR_MACRO = """
from karabo.middlelayer import Macro123, Slot, String

class InvalidMacro(Macro)
    name = String(defaultValue="Foo")

    @Slot()
    def execute(self):
        print("Hello {{}}!".format(name))
"""

INDENTATION_ERROR_MACRO = """
from karabo.middlelayer import Macro, Slot, String

  class InvalidMacro(Macro)
    name = String(defaultValue="Foo")

    @Slot()
    def execute(self):
        print("Hello {{}}!".format(name))
"""



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from uuid import uuid4

from qtpy.QtNetwork import QAbstractSocket
from qtpy.QtWidgets import QDialog

import karabogui.access as krb_access
from karabo.native import Hash
from karabogui.singletons.configuration import Configuration
from karabogui.singletons.mediator import Mediator
from karabogui.singletons.network import Network
from karabogui.testing import singletons


class MockReceiver:
    def __init__(self):
        self.last_hash = None
        self.server_connection = False

    def slotReceiveData(self, data):
        self.last_hash = data

    def slotServerConnection(self, status):
        self.server_connection = status


def test_member_variables(gui_app):
    network = Network()
    with singletons(network=network):
        assert network.hostname == 'localhost'
        assert network.password == 'karabo'
        assert network.port == 44444


def test_connect_directly(mocker, gui_app):
    socket = mocker.patch('karabogui.singletons.network.QTcpSocket')
    network = Network()
    config = Configuration()
    with singletons(network=network, configuration=config):
        host = "xfel-computer-system-control"
        port = 33333
        mock_login_dialog = mocker.patch(
            "karabogui.singletons.network.ReactiveLoginDialog")
        mock_login_dialog().hostname = host
        mock_login_dialog().port = port
        mock_login_dialog().gui_servers = []
        mock_login_dialog().access_level = "operator"
        mock_login_dialog().exec.return_value = QDialog.Accepted
        success = network.connectToServerDirectly(host, port)
        assert success
        socket().connectToHost.assert_called_with(host, port)

        network.onConnected()
        # And disconnect!
        network.disconnectFromServer()
        socket().disconnectFromHost.assert_called_once()


def test_socket_connect_login_protocol(mocker, subtests, gui_app):
    socket = mocker.patch('karabogui.singletons.network.QTcpSocket')
    network = Network()
    mediator = Mediator()
    receiver = MockReceiver()
    network.signalReceivedData.connect(receiver.slotReceiveData)
    network.signalServerConnectionChanged.connect(
        receiver.slotServerConnection)
    with singletons(network=network, mediator=mediator):
        host = "exfl-client-pc-system-guikarabo"
        port = 32323
        mock_login_dialog = mocker.patch(
            "karabogui.singletons.network.ReactiveLoginDialog")
        mock_login_dialog().hostname = host
        mock_login_dialog().port = port
        mock_login_dialog().gui_servers = []
        mock_login_dialog().access_level = "operator"

        mock_login_dialog().exec.return_value = QDialog.Accepted
        success = network.connectToServerDirectly(host, port)
        assert success
        socket().connectToHost.assert_called_with(host, port)

        # We received reply, we set server information and connect
        # send our login information!
        network.set_server_information(read_only=False)

        def _trigger_message_parse():
            nonlocal socket
            # Get the byte array and remove size
            call = socket().write.call_args[0]
            byte_array = call[0]
            byte_array = byte_array[4:]
            data = byte_array.data()
            # Place timestamp to await and parse Input to trigger signal
            network._waiting_messages[id(data)] = 0
            network.parseInput(data)

        _trigger_message_parse()
        assert receiver.last_hash is not None
        assert isinstance(receiver.last_hash, Hash)
        assert receiver.last_hash["type"] == "login"

        with subtests.test(msg="Test alarm interface"):
            instanceId = "test"
            network.onAcknowledgeAlarm(instanceId=instanceId, rowId="2")
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "acknowledgeAlarm"
            assert receiver.last_hash["alarmInstanceId"] == instanceId
            assert receiver.last_hash["acknowledgedRows"] == Hash("2", True)

            network.onRequestAlarms(instanceId)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestAlarms"
            assert receiver.last_hash["alarmInstanceId"] == instanceId

        with subtests.test(msg="Test misc network"):
            output_device = "NODEVICE/DESY/GLASPALACE"
            network.onSubscribeToOutput(output_device, "output", True)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "subscribeNetwork"
            assert receiver.last_hash[
                       "channelName"] == f"{output_device}:output"
            assert receiver.last_hash["subscribe"] is True

            network.onRequestNetwork(f"{output_device}:output")
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestNetwork"
            assert receiver.last_hash["channelName"] == \
                   f"{output_device}:output"

            text = "This is an error"
            network.onInfo(Hash("traceback", text, "type", "error"))
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "error"
            assert receiver.last_hash["info"]["traceback"] == text

        with subtests.test(msg="Test Network Project Interface"):
            project_manager = "KaraboProjectDB"

            network.onListProjectDomains(project_manager)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["instanceId"] == project_manager

            network.onProjectListItems(project_manager, "FXE", "scene")
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["instanceId"] == project_manager
            assert receiver.last_hash["args.domain"] == "FXE"
            assert receiver.last_hash["args.item_types"] == ["scene"]

            items = [str(uuid4()), str(uuid4())]
            network.onProjectLoadItems(project_manager, items)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["instanceId"] == project_manager
            assert receiver.last_hash["args.items"] == items

            network.onProjectSaveItems(project_manager, items)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["instanceId"] == project_manager
            assert receiver.last_hash["args.items"] == items
            assert receiver.last_hash["args.client"] is not None

            network.onProjectUpdateTrashed(project_manager, items)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["instanceId"] == project_manager
            assert receiver.last_hash["args.items"] == items

        with subtests.test(msg="Protocol Methods"):
            deviceId = "NoDeviceHere"
            network.onKillDevice(deviceId)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "killDevice"
            assert receiver.last_hash["deviceId"] == deviceId

            serverId = "ProjectDBServer"
            network.onKillServer(serverId)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "killServer"
            assert receiver.last_hash["serverId"] == serverId

            network.onGetDeviceConfiguration(deviceId)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "getDeviceConfiguration"
            assert receiver.last_hash["deviceId"] == deviceId

            dev_config = Hash("position", 5)
            network.onReconfigure(deviceId, dev_config)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "reconfigure"
            assert receiver.last_hash["deviceId"] == deviceId
            assert receiver.last_hash["configuration"] == dev_config
            assert receiver.last_hash["reply"] is True
            assert receiver.last_hash["timeout"] == 5

            classId = "NoSpecialClassId"
            network.onInitDevice(serverId, classId, deviceId, dev_config)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "initDevice"
            assert receiver.last_hash["deviceId"] == deviceId
            assert receiver.last_hash["serverId"] == serverId
            assert receiver.last_hash["classId"] == classId
            assert receiver.last_hash["configuration"] == dev_config

            slotName = "start"
            network.onExecute(deviceId, slotName, False)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "execute"
            assert receiver.last_hash["deviceId"] == deviceId
            assert receiver.last_hash["command"] == slotName
            assert receiver.last_hash["reply"] is True
            assert receiver.last_hash["timeout"] == 5

            network.onExecuteGeneric(deviceId, slotName, dev_config)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["instanceId"] == deviceId
            assert receiver.last_hash["slot"] == slotName
            assert receiver.last_hash["args"] == dev_config
            assert receiver.last_hash["timeout"] == 5
            assert receiver.last_hash["replyType"] == "requestGeneric"

            network.onStartMonitoringDevice(deviceId)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "startMonitoringDevice"
            assert receiver.last_hash["deviceId"] == deviceId

            network.onStopMonitoringDevice(deviceId)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "stopMonitoringDevice"
            assert receiver.last_hash["deviceId"] == deviceId

            network.onGetClassSchema(serverId, classId)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "getClassSchema"
            assert receiver.last_hash["serverId"] == serverId
            assert receiver.last_hash["classId"] == classId

            network.onGetDeviceSchema(deviceId)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "getDeviceSchema"
            assert receiver.last_hash["deviceId"] == deviceId

            path = "position"
            t0 = 0
            t1 = 1
            maxNumData = 500
            network.onGetPropertyHistory(deviceId, path, t0, t1,
                                         maxNumData)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "getPropertyHistory"
            assert receiver.last_hash["deviceId"] == deviceId
            assert receiver.last_hash["t0"] == t0
            assert receiver.last_hash["t1"] == t1
            assert receiver.last_hash["property"] == path
            assert receiver.last_hash["maxNumData"] == maxNumData

        with subtests.test(msg="Configuration Interface"):
            preview = True
            network.onGetConfigurationFromPast(deviceId, t0, preview)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "getConfigurationFromPast"
            assert receiver.last_hash["deviceId"] == deviceId
            assert receiver.last_hash["time"] == t0
            assert receiver.last_hash["preview"] == preview

            conf_name = "test"
            network.onGetInitConfiguration(deviceId, conf_name,
                                           preview)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["args.deviceId"] == deviceId
            assert receiver.last_hash["args.name"] == conf_name
            assert receiver.last_hash["timeout"] == 5
            assert receiver.last_hash["preview"] == preview
            assert receiver.last_hash[
                       "slot"] == "slotGetInitConfiguration"
            assert receiver.last_hash[
                       "replyType"] == "getInitConfiguration"

            conf_name = "test2"
            network.onDeleteInitConfiguration(deviceId, conf_name)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["args.deviceId"] == deviceId
            assert receiver.last_hash["args.name"] == conf_name
            assert receiver.last_hash["timeout"] == 5
            assert receiver.last_hash[
                       "slot"] == "slotDeleteInitConfiguration"
            assert receiver.last_hash[
                       "replyType"] == "deleteInitConfiguration"

            network.onListInitConfigurations(deviceId)
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["args.deviceId"] == deviceId
            assert receiver.last_hash["args.name"] == ""
            assert receiver.last_hash[
                       "slot"] == "slotListInitConfigurations"
            assert receiver.last_hash[
                       "replyType"] == "listInitConfigurations"

            network.onSaveInitConfiguration(conf_name, [deviceId])
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["args.deviceIds"] == [deviceId]
            assert receiver.last_hash["args.name"] == conf_name
            assert receiver.last_hash[
                       "slot"] == "slotSaveInitConfiguration"
            assert receiver.last_hash[
                       "replyType"] == "saveInitConfiguration"
            assert receiver.last_hash["timeout"] == 5

        with subtests.test(msg="Log messages"):
            network.onSetLogLevel("swerver", "ERROR")
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "setLogLevel"
            assert receiver.last_hash["instanceId"] == "swerver"
            assert receiver.last_hash["level"] == "ERROR"

        with subtests.test(msg="Info"):
            network.onGetGuiSessionInfo()
            _trigger_message_parse()
            assert receiver.last_hash["type"] == "getGuiSessionInfo"

        mbox = mocker.patch('karabogui.singletons.network.QMessageBox')
        dia = mocker.patch('karabogui.singletons.network.ReactiveLoginDialog')
        call_count = 0
        for error in [QAbstractSocket.ConnectionRefusedError,
                      QAbstractSocket.RemoteHostClosedError,
                      QAbstractSocket.HostNotFoundError,
                      QAbstractSocket.NetworkError,
                      QAbstractSocket.SocketAccessError,
                      QAbstractSocket.DatagramTooLargeError]:
            # Last error not in list of expected !
            network.onSocketError(error)
            call_count += 1
            mbox.question.call_count = call_count
            # We are asked to relogin on error
            dia.call_count = call_count

        with subtests.test(msg="Active Proposals"):
            network.listDestinations()
            _trigger_message_parse()
            h = Hash("type", "requestGeneric",
                     "instanceId", "KaraboLogBook",
                     "slot", "slotListDestinations",
                     "replyType", "listDestinations",
                     "args", Hash(),
                     )
            assert receiver.last_hash.fullyEqual(h)

        with subtests.test(msg="SaveLogBook"):
            name = "TestProposal"
            dataType = "Image"
            data = "Test data"
            caption = "This is a test message"
            network.onSaveLogBook(
                name=name, dataType=dataType, data=data,
                caption=caption)

            _trigger_message_parse()
            assert receiver.last_hash["type"] == "requestGeneric"
            assert receiver.last_hash["instanceId"] == "KaraboLogBook"
            assert receiver.last_hash["args.name"] == name
            assert receiver.last_hash["args.dataType"] == dataType

        with subtests.test(msg="Network connection"):
            assert receiver.server_connection is False
            network.onConnected()
            assert receiver.server_connection is True
            krb_access.ONE_TIME_TOKEN = 123345
            krb_access.TEMPORARY_SESSION_USER = "karabo"
            krb_access.TEMPORARY_SESSION_WARNING = True
            network.disconnectFromServer()
            assert receiver.server_connection is False
            assert krb_access.ONE_TIME_TOKEN is None
            assert krb_access.TEMPORARY_SESSION_USER is None
            assert krb_access.TEMPORARY_SESSION_WARNING is False



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os
import sys

import pytest
from qtpy.QtWidgets import QApplication

from karabo.native import AccessLevel
from karabogui.background import create_background_timer
from karabogui.controllers.api import populate_controller_registry


@pytest.fixture(scope="package")
def gui_app():
    os.environ["KARABO_TEST_GUI"] = "1"
    app = QApplication.instance()
    if app is None:
        app = QApplication(sys.argv)
    # Check organization for QSettings for tests, see the setting in
    # `programs/tests/test_base.py`
    app.setOrganizationName("TestFacility")
    app.setOrganizationDomain("karabo.test.eu")
    app.setApplicationName("pytest")
    assert app.applicationName() == "pytest"
    create_background_timer()
    import karabogui.access as krb_access
    krb_access.GLOBAL_ACCESS_LEVEL = AccessLevel.OPERATOR
    from karabogui import icons
    icons.init()
    populate_controller_registry()
    return app



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import pytest
from qtpy.QtCore import QEvent, QPoint, Qt
from qtpy.QtGui import QMouseEvent
from qtpy.QtWidgets import QWidget
from traits.api import Dict, Instance

from karabo.common.scenemodel.api import BaseWidgetObjectData, SceneModel
from karabo.native import Configurable, Double
from karabogui import icons
from karabogui.binding.api import FloatBinding, ProxyStatus, StringBinding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.sceneview.api import ControllerContainer, SceneView
from karabogui.sceneview.tools.widgethandler import SceneControllerHandler
from karabogui.testing import flushed_registry, singletons, system_hash
from karabogui.topology.api import SystemTopology


class TypeObject(Configurable):
    prop = Double(
        minInc=-1000.0,
        maxInc=1000.0)


class UniqueModel(BaseWidgetObjectData):
    attributes = Dict


@pytest.fixture(scope="module")
def display_widget():
    with flushed_registry():
        @register_binding_controller(
            klassname="Editor",
            binding_type=(StringBinding, FloatBinding))
        class DisplayWidget(BaseBindingController):
            model = Instance(UniqueModel)

            @staticmethod
            def initialize_model(proxy, model):
                model.attributes = proxy.binding.attributes

            def create_widget(self, parent=None):
                return QWidget(parent)

        yield DisplayWidget


def test_scene_widget_handler(gui_app, display_widget, mocker):
    """Test the scene widget handler"""
    topology = SystemTopology()
    topology.initialize(system_hash())
    with singletons(topology=topology):
        scene_model = SceneModel(children=[])
        scene_view = SceneView(model=scene_model)
        models = UniqueModel(
            keys=["divyy.prop"],
            attributes={"One": "Two"})
        scene_view.add_models(models)

        assert len(scene_model.children) == 1

        item = scene_view.controller_at_position(QPoint(0, 0))
        assert item is not None
        assert isinstance(item, ControllerContainer)
        assert isinstance(item.widget_controller, display_widget)
        assert "minInc" not in item.model.attributes
        assert "maxInc" not in item.model.attributes

        handler = SceneControllerHandler(widget=item)
        assert handler.can_handle()

        event = QMouseEvent(
            QEvent.MouseButtonPress,
            QPoint(80, 0),
            Qt.RightButton,
            Qt.RightButton,
            Qt.NoModifier)

        menu = mocker.MagicMock()
        handler.handle_widget(scene_view, event, menu)
        menu.addAction.assert_called_with("No mutation for offline properties")

        item.widget_controller.proxy.root_proxy.status = ProxyStatus.ONLINE
        topology.device_schema_updated("divyy", TypeObject.getClassSchema())
        assert item.widget_controller.proxy.binding is not None

        menu.reset_mock()
        handler.handle_widget(scene_view, event, menu)
        menu.addMenu.assert_called_with(icons.change, "Change Widget")

        # Change the widget
        handler._change_widget(scene_view, display_widget)
        item = scene_view.controller_at_position(QPoint(0, 0))
        assert item is not None
        assert isinstance(item, ControllerContainer)
        assert isinstance(item.widget_controller, display_widget)
        assert item.model.attributes["minInc"] == -1000.0
        assert item.model.attributes["maxInc"] == 1000.0

        scene_view.destroy()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtWidgets import QAction, QDialog
from traits.api import Instance, Str, on_trait_change

from karabo.common.scenemodel.api import DisplayStateColorModel
from karabogui.binding.api import StringBinding
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller, with_display_type)
from karabogui.dialogs.format_label import FormatLabelDialog
from karabogui.fonts import get_font_size_from_dpi
from karabogui.indicators import get_state_color
from karabogui.util import generateObjectName
from karabogui.widgets.hints import FrameWidget


@register_binding_controller(ui_name='State Color Field',
                             klassname='DisplayStateColor',
                             binding_type=StringBinding,
                             priority=90,
                             is_compatible=with_display_type('State'),
                             can_show_nothing=False)
class DisplayStateColor(BaseBindingController):
    # The specific scene model class used by this widget
    model = Instance(DisplayStateColorModel, args=())
    # CSS template for the widget BG color
    _style_sheet = Str

    def create_widget(self, parent):
        widget = FrameWidget(parent)
        objectName = generateObjectName(self)
        self._style_sheet = (f"QLabel#{objectName}" +
                             " {{ background-color : rgb{}; }}")
        widget.setObjectName(objectName)

        textAction = QAction("Show State String", widget)
        textAction.triggered.connect(self._show_state_string)
        # update the context menu and keep track
        textAction.setCheckable(True)
        textAction.setChecked(self.model.show_string)
        widget.addAction(textAction)

        # Add an action for formatting options
        format_action = QAction("Format field..", widget)
        format_action.triggered.connect(self._format_field)
        widget.addAction(format_action)
        self._apply_format(widget)

        return widget

    def value_update(self, proxy):
        value = proxy.value
        color = get_state_color(value)
        sheet = self._style_sheet.format(color)
        self.widget.setStyleSheet(sheet)

        if self.model.show_string:
            self.widget.setText(proxy.value)

    @on_trait_change('model.show_string', post_init=True)
    def _update_text(self):
        if self.proxy is not None:
            self.value_update(self.proxy)
            if not self.model.show_string:
                # Only clear the widget once if no action is set!
                self.widget.clear()

    def _show_state_string(self):
        self.model.show_string = not self.model.show_string

    # -----------------------------------------------------------------------
    # Formatting methods

    def _format_field(self):
        dialog = FormatLabelDialog(font_size=self.model.font_size,
                                   font_weight=self.model.font_weight,
                                   parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            self.model.trait_set(font_size=dialog.font_size,
                                 font_weight=dialog.font_weight)
            self._apply_format()

    def _apply_format(self, widget=None):
        """The widget is passed as an argument in create_widget as it is not
           yet bound to self.widget then"""
        if widget is None:
            widget = self.widget

        # Apply font formatting
        font = widget.font()
        font.setPointSize(get_font_size_from_dpi(self.model.font_size))
        font.setBold(self.model.font_weight == "bold")
        widget.setFont(font)



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on May 21, 2022
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from datetime import datetime, timezone

import numpy as np
from qtpy.QtCore import QDateTime, QSortFilterProxyModel, QStringListModel, Qt
from qtpy.QtGui import QFont
from qtpy.QtWidgets import (
    QAbstractItemView, QAction, QComboBox, QDialog, QFrame, QHBoxLayout,
    QLabel, QLineEdit, QListView, QMenu, QPushButton, QTextEdit, QVBoxLayout,
    QWidget)
from traits.api import Instance, Int, String, WeakRef, on_trait_change

from karabo.common.scenemodel.api import HistoricTextModel
from karabo.native import Timestamp
from karabogui import icons
from karabogui.binding.api import (
    StringBinding, UnsignedIntBinding, VectorStringBinding, get_binding_value,
    get_numpy_binding)
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.dialogs.api import RequestTimeDialog
from karabogui.indicators import ALL_OK_COLOR
from karabogui.util import generateObjectName

# Time Ranges
LAST_WEEK = "Last Week"
LAST_DAY = "Last Day"
LAST_HOUR = "Last Hour"
LAST_MIN = "Last Ten Minutes"


def get_formatted(ts):
    stamp = (f"[{ts.day:02}-{ts.month:02}-{ts.year} "
             f"{ts.hour:02}:{ts.minute:02}:"
             f"{ts.second:02}.{ts.microsecond // 1000:03d}]")
    return stamp


def get_dtype(binding):
    """Return a dtype corresponding to binding"""
    fmt = "{}"
    if isinstance(binding, UnsignedIntBinding):
        dt = binding.displayType.split("|")[0]
        _fmt = {
            "bin": "0b{{:0{}b}}".format(
                np.iinfo(get_numpy_binding(binding)).bits),
            "oct": "o{:o}",
            "hex": "0x{:X}"}
        try:
            fmt = _fmt[dt]
        except (TypeError, KeyError):
            pass

    return fmt


def get_start_end_date_time(time_span):
    """ Return beginning and end time for given ``time_span``

    :returns: tuple (float, float) of start, end time.
    """
    current_date_time = QDateTime.currentDateTime()
    if time_span == LAST_WEEK:
        # One week
        start_date_time = current_date_time.addDays(-7)
    if time_span == LAST_DAY:
        # One day
        start_date_time = current_date_time.addDays(-1)
    elif time_span == LAST_HOUR:
        # One hour
        start_date_time = current_date_time.addSecs(-3600)
    elif time_span == LAST_MIN:
        # Ten minutes
        start_date_time = current_date_time.addSecs(-600)

    start_time = start_date_time.toMSecsSinceEpoch() / 1000
    current_time = current_date_time.toMSecsSinceEpoch() / 1000

    return start_time, current_time


_BINDINGS = (UnsignedIntBinding, StringBinding, VectorStringBinding)


@register_binding_controller(ui_name="Historic Text Data", can_edit=False,
                             klassname="HistoricText",
                             binding_type=_BINDINGS,
                             priority=-10)
class DisplayHistoricText(BaseBindingController):
    """
    The HistoricText Controller for viewing historic data of strings and
    vectors of strings
    """
    model = Instance(HistoricTextModel, args=())

    list_model = WeakRef(QStringListModel)
    list_view = WeakRef(QListView)
    time_span = WeakRef(QComboBox)
    status_widget = WeakRef(QLabel)
    request_button = WeakRef(QPushButton)
    text_widget = WeakRef(QTextEdit)

    maxHistory = Int(500)
    fmt = String("{}")

    def create_widget(self, parent):
        widget = QWidget(parent)
        layout = QVBoxLayout(widget)
        widget.setLayout(layout)

        bold_font = QFont()
        bold_font.setPointSize(10)
        bold_font.setBold(True)

        text_label = QLabel("Current Value")
        text_label.setFont(bold_font)

        text_widget = QTextEdit(widget)
        text_widget.setFocusPolicy(Qt.NoFocus)
        text_widget.setFixedHeight(60)
        text_widget.setAlignment(Qt.AlignCenter)
        objectName = generateObjectName(widget)
        sheet = ('QWidget#{} {{ background-color : rgba{}; }}'
                 ''.format(objectName, ALL_OK_COLOR))
        text_widget.setObjectName(objectName)
        text_widget.setStyleSheet(sheet)

        history_label = QLabel("Historic Values")
        history_label.setFont(bold_font)

        list_view = QListView(widget)
        list_view.setWordWrap(True)
        list_view.setAlternatingRowColors(True)
        list_view.setEditTriggers(QAbstractItemView.NoEditTriggers)
        self.list_model = QStringListModel(parent=list_view)

        # Set up the filter model!
        filter_model = QSortFilterProxyModel(parent=list_view)
        filter_model.setSourceModel(self.list_model)
        filter_model.setFilterRole(Qt.DisplayRole)
        filter_model.setFilterCaseSensitivity(False)
        filter_model.setFilterFixedString("")
        list_view.setModel(filter_model)

        search_layout = QHBoxLayout()
        search_layout.setContentsMargins(0, 0, 0, 0)
        search_line = QLineEdit(parent=widget)
        clear_button = QPushButton("Clear Filter", parent=widget)
        clear_button.clicked.connect(search_line.clear)
        search_line.textChanged.connect(filter_model.setFilterFixedString)

        search_layout.addWidget(search_line)
        search_layout.addWidget(clear_button)

        button_layout = QHBoxLayout()
        button_layout.setContentsMargins(0, 0, 0, 0)

        time_span = QComboBox()
        time_span.addItems([LAST_MIN, LAST_HOUR, LAST_DAY, LAST_WEEK])
        request_button = QPushButton("Request History")
        request_button.clicked.connect(self._get_history)
        button_layout.addWidget(time_span)
        button_layout.addWidget(request_button)

        status_widget = QLabel()
        status_widget.setFont(bold_font)
        status_widget.setFrameShape(QFrame.Box)

        layout.addWidget(text_label)
        layout.addWidget(text_widget)
        layout.addWidget(history_label)
        layout.addLayout(search_layout)
        layout.addWidget(list_view)
        layout.addLayout(button_layout)
        layout.addWidget(status_widget)

        self.list_view = list_view
        self.time_span = time_span
        self.status_widget = status_widget
        self.request_button = request_button
        self.text_widget = text_widget
        widget.setContextMenuPolicy(Qt.CustomContextMenu)
        widget.customContextMenuRequested.connect(self._context_menu)
        self._write_status("Historic text widget created ...")

        return widget

    def binding_update(self, proxy):
        self.fmt = get_dtype(proxy.binding)

    def value_update(self, proxy):
        value = get_binding_value(proxy)
        if value is None:
            return
        timestamp = proxy.binding.timestamp
        dt = datetime.fromtimestamp(timestamp.toTimestamp())
        stamp = get_formatted(dt)
        value = self._create_value(value)
        self.text_widget.setText(f"{stamp} {value}")

    # ----------------------------------------------------------------------
    # Internal interface

    def _create_value(self, value):
        """Transform the value"""
        try:
            return self.fmt.format(value)
        except BaseException:
            return str(value)

    def _convert_datetime_utc(self, start, end):
        """Convert datetime objects for `start` and `end` to UTC strings"""
        start = str(datetime.fromtimestamp(start, tz=timezone.utc).isoformat())
        end = str(datetime.fromtimestamp(end, tz=timezone.utc).isoformat())
        return start, end

    @on_trait_change("proxy:binding:historic_data")
    def _historic_data_arrival(self, data):
        """Historic data arrived, handle it"""
        size = len(data)
        self._write_status(f"Received historic data with {size} entries.")

        def format_entry(h):
            """Format an entry of the historic data"""
            timestamp = Timestamp.fromHashAttributes(h["v", ...])
            dt = datetime.fromtimestamp(timestamp.toTimestamp())
            stamp = get_formatted(dt)
            value = self._create_value(h["v"])
            return f"{stamp} {value}"

        data = [format_entry(h) for h in data][::-1]
        self.list_model.setStringList(data)

    # -----------------------------------------------------------------------
    # Slots

    def _context_menu(self, pos):
        menu = QMenu(parent=self.widget)
        dialog_ac = QAction(icons.clock, "Request Time", parent=self.widget)
        dialog_ac.triggered.connect(self._request_dialog)
        menu.addAction(dialog_ac)

        menu.exec(self.widget.mapToGlobal(pos))

    def _request_dialog(self):
        """Set a new time interval and request historic data"""
        start, end = QDateTime.currentDateTime(), QDateTime.currentDateTime()
        dialog = RequestTimeDialog(start=start, end=end, parent=self.widget)
        if dialog.exec() == QDialog.Accepted:
            start, end = self._convert_datetime_utc(
                *dialog.get_start_and_end_time())
            self.proxy.get_history(start, end, max_value_count=self.maxHistory)

    def _get_history(self):
        """Send a historic data request to the dataloggers"""
        start, end = self._convert_datetime_utc(
            *get_start_end_date_time(self.time_span.currentText()))
        self.proxy.get_history(start, end, max_value_count=self.maxHistory)
        self._write_status("Requesting historic data!")

    def _write_status(self, text):
        """Write a status on the label with the time point"""
        dt = datetime.fromtimestamp(Timestamp().toTimestamp())
        stamp = get_formatted(dt)
        self.status_widget.setText(f"{stamp}: {text}")



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy import uic
from qtpy.QtCore import Slot
from qtpy.QtWidgets import QDialog

from karabo.common.scenemodel.const import SCENE_FONT_SIZES, SCENE_FONT_WEIGHTS

from .utils import get_dialog_ui


class FormatLabelDialog(QDialog):

    def __init__(self, font_size=10, font_weight="normal", parent=None):
        super().__init__(parent=parent)
        uic.loadUi(get_dialog_ui("format_label.ui"), self)

        # Populate font size combobox and set current index
        sizes_string = [str(size) for size in SCENE_FONT_SIZES]
        self.font_size_combobox.addItems(sizes_string)
        index = SCENE_FONT_SIZES.index(font_size)
        self.font_size_combobox.setCurrentIndex(index)

        # Populate font weight combobox and set current index
        self.font_weight_combobox.addItems(SCENE_FONT_WEIGHTS)
        index = SCENE_FONT_WEIGHTS.index(font_weight)
        self.font_weight_combobox.setCurrentIndex(index)

        # Add triggers to change preview
        self.font_size_combobox.currentIndexChanged.connect(self._preview)
        self.font_weight_combobox.currentIndexChanged.connect(self._preview)

        # Update preview to reflect current format
        self._preview()

    @Slot()
    def _preview(self):
        font = self.preview_label.font()
        font.setPointSize(self.font_size)
        font.setBold(self.font_weight == "bold")
        self.preview_label.setFont(font)

    @property
    def font_size(self):
        return int(self.font_size_combobox.currentText())

    @property
    def font_weight(self):
        return self.font_weight_combobox.currentText()



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QSize, Qt, Signal, Slot
from qtpy.QtWidgets import QAction, QMenu, QToolBar, QToolButton, QWidgetAction

from ..const import BUTTON_SIZE, ICON_SIZE


class ToolBar(QToolBar):

    def __init__(self, orientation=Qt.Vertical, parent=None):
        super().__init__(parent)
        # Setup toolbar settings
        self.setOrientation(orientation)
        self.setStyleSheet("QToolBar { border: 0px }")
        self._separators = {}

    def add_button(self, button):
        """Wraps the button into an action to have seamless integration
           with the toolbar"""
        action = WidgetAction(button, parent=self)

        # Check separators
        self.addAction(action)

    def setBackground(self, color):
        sheet = "QToolBar {{ border: 0px; background-color: rgba{}; }}"
        self.setStyleSheet(sheet.format(color.getRgb()))


class WidgetAction(QWidgetAction):
    def __init__(self, button, parent=None):
        super().__init__(parent)
        button.setParent(parent)
        self._toolbar_button = button
        self._menu_button = None
        self._actions = {}

    def createWidget(self, parent):
        button = self._toolbar_button
        if parent is self.parent():
            # The parent is the toolbar. Return the original button, which
            # is to be added on the toolbar itself.
            return button

        # The parent is the extension menu. We create a tool button
        # that emulates the toolbar button

        # Distinguish what to create (normal button vs. dropdown button)
        menu = button.menu()
        if menu:
            # Copy all the actions from the dropdown menu and bookkeep them
            actions = []
            for action in menu.actions():
                new_action = create_action(text=action.text(),
                                           icon=action.icon(),
                                           data=action.data(),
                                           checkable=action.isCheckable())
                if action.isCheckable():
                    new_action.setChecked(action.isChecked())

                actions.append(new_action)
                self._actions[action] = new_action

            # Create a dropdown button using the information from the
            # toolbar button
            new = create_dropdown_button(
                actions=actions,
                icon=button.icon(),
                tooltip=button.toolTip(),
                checkable=button.isCheckable(),
                on_clicked=self._on_menu_button_clicked)

            # Set the corresponding copied default action
            new.setDefaultAction(self._actions[button.defaultAction()])

            # Intercept the menu signal, we update the copied button according
            # to the logic of the original button
            parent.aboutToShow.connect(self._update_menu_button)
        else:
            # Create a simple button. Not as fancy as the other one.
            new = create_button(
                icon=button.icon(),
                checkable=button.isCheckable(),
                tooltip=button.toolTip(),
                on_clicked=self._on_menu_button_clicked)

        # Proper parenting is required.
        new.setParent(parent)

        # Set checked state
        if button.isCheckable():
            new.setChecked(button.isChecked())

        # We hide the menu every when button is clicked.
        new.clicked.connect(parent.hide)

        self._menu_button = new
        return new

    def deleteWidget(self, widget):
        # Clean the menu signal
        if widget.menu():
            parent = widget.parent()
            if isinstance(parent, QMenu):
                parent.aboutToShow.disconnect(self._update_menu_button)
            self._actions.clear()

        self._menu_button = None
        super().deleteWidget(widget)

    @Slot()
    def _on_menu_button_clicked(self):
        # Change the default action according to the selected action
        for orig_action, new_action in self._actions.items():
            if new_action is self._menu_button.defaultAction():
                self._toolbar_button.setDefaultAction(orig_action)

        self._toolbar_button.click()

    def _update_menu_button(self):
        orig_action = self._toolbar_button.defaultAction()
        new_action = self._actions[orig_action]
        self._menu_button.setDefaultAction(new_action)
        new_action.setChecked(orig_action.isChecked())


class DropDownMenu(QMenu):
    triggered = Signal(QAction)

    def mouseReleaseEvent(self, event):
        super().mouseReleaseEvent(event)
        if event.button() == Qt.LeftButton:
            action = self.actionAt(event.pos())
            if action:
                self.triggered.emit(action)


class DropDownButton(QToolButton):
    # A similar signal with `clicked` but contains the data from the actions.
    # Will return `None` if unchecked.
    triggered = Signal(object)

    def __init__(self, parent=None):
        super().__init__(parent=parent)

        # Setup custom dropdown menu
        menu = DropDownMenu()
        menu.triggered.connect(self._set_default_action)
        self.setMenu(menu)
        self.setPopupMode(QToolButton.MenuButtonPopup)

        # Connect signal
        self.clicked.connect(self._trigger_data)

    def set_actions(self, actions):
        menu = self.menu()
        for action in actions:
            # Action should be checkable to accept button check state
            action.setParent(menu)
            menu.addAction(action)
        self._set_default_action(actions[0], enabled=False)

    def check(self, data):
        for action in self.menu().actions():
            if data == action.data():
                self.setDefaultAction(action)
                self.setChecked(True)
                # Bail out if we have find the action
                return

        self.setChecked(False)

    def setCheckable(self, checkable):
        super().setCheckable(checkable)
        if checkable:
            self.toggled.connect(self._check_action)

    @Slot(bool)
    def _trigger_data(self, checked):
        data = None
        if checked:
            data = self.defaultAction().data()
        self.triggered.emit(data)

    @Slot(bool)
    def _check_action(self, checked):
        current = self.defaultAction()
        for action in self.menu().actions():
            if action is not current:
                action.setChecked(False)
        current.setChecked(checked)

    def _set_default_action(self, action, enabled=True):
        """Callback when select"""
        # Reset action check state, will be rechecked by button click.
        if self.isCheckable():
            action.setChecked(False)
        self.setDefaultAction(action)

        # Click immediately upon selected
        if enabled:
            self.click()


def create_button(*, checkable, icon, tooltip, on_clicked=None):
    button = QToolButton()
    button.setCheckable(checkable)
    button.setIcon(icon)
    button.setAutoRaise(True)
    button.setToolTip(tooltip)
    button.setIconSize(QSize(ICON_SIZE, ICON_SIZE))
    button.setFixedSize(BUTTON_SIZE, BUTTON_SIZE)
    if on_clicked is not None:
        button.clicked.connect(on_clicked)

    return button


# XXX: Backward compatibility
create_tool_button = create_button


def create_dropdown_button(actions, checkable, icon, tooltip,
                           on_clicked=None):
    button = DropDownButton()
    button.setCheckable(checkable)
    button.setIcon(icon)
    button.setAutoRaise(True)
    button.setToolTip(tooltip)
    button.setIconSize(QSize(ICON_SIZE, ICON_SIZE))
    button.setFixedSize(BUTTON_SIZE, BUTTON_SIZE)
    if on_clicked is not None:
        button.clicked.connect(on_clicked)

    # Main difference from the normal tool button: actions as children
    button.set_actions(actions)

    return button


def create_action(text, icon, data=None, on_clicked=None, checkable=False):
    action = QAction(text, None)
    action.setIcon(icon)
    action.setCheckable(checkable)
    if data is not None:
        action.setData(data)
    if on_clicked is not None:
        action.triggered.connect(on_clicked)

    return action



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from qtpy.QtCore import QRect
from qtpy.QtWidgets import QWidget

from karabo.common.scenemodel.api import StickerModel
from karabogui.sceneview.widget.sticker import StickerDialog, StickerWidget
from karabogui.testing import GuiTestCase

TEXT = """
Lorem ipsum dolor sit amet, consetetur sadipscing elitr, sed diam nonumy eirmod
tempor invidunt ut labore et dolore magna aliquyam erat, sed diam voluptua. At
vero eos et accusam et justo duo dolores et ea rebum. Stet clita kasd
gubergren, no sea takimata sanctus est Lorem ipsum dolor sit amet."""


class TestStickerWidget(GuiTestCase):

    """Test for the Sticker Widget is limited as this relies heavily on
       CSS customizations. The only way to test these values is just by
       intercepting the return values from a mocked dialog"""

    # -----------------------------------------------------------------------
    # Test setup

    def setUp(self):
        super().setUp()
        self.model = model = StickerModel(text=TEXT, foreground='black',
                                          x=0, y=0, width=100, height=100)

        # The sticker widget needs to have a shown parent in order to
        # test the sizes.
        self.parent = QWidget()
        self.widget = StickerWidget(model=model, parent=self.parent)
        self.parent.show()

    def tearDown(self):
        super().tearDown()
        self.parent = None
        self.widget = None

    # -----------------------------------------------------------------------
    # Actual tests

    def test_basics(self):
        # Check widget properties
        assert self.widget.toPlainText() == self.model.text
        model_rect = QRect(self.model.x, self.model.y,
                           self.model.width, self.model.height)
        assert self.widget.rect() == model_rect

        # Check dialog properties. It is needed to be shown to process painting
        # and check sizes
        dialog = StickerDialog(model=self.model, parent=self.widget)
        dialog.show()
        dialog_sticker = dialog.leText

        assert dialog_sticker.size() == self.widget.size()
        assert dialog_sticker.toPlainText() == self.model.text

        are_viewports_equivalent = (dialog_sticker.viewport().size()
                                    == self.widget.viewport().size())
        assert are_viewports_equivalent



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on November 23, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import importlib

from qtpy.QtWidgets import QApplication

# module global for storing singleton instances
__singletons = {}


def _build_getter(key, modulename, klassname):
    """A factory function for building singleton getter functions

    This is an internal detail, but some explanation would be helpful.
    Basically, the pattern for instantiating a singleton is to import a class
    from some module, then instantiate and store the result in a module global
    which is then accessible from some module-level getter function.

    The generalization of that is this: We have a single module-level dict
    which contains our singleton instances named ``__singletons``. For each
    singleton we define a getter function which checks for the existence of a
    specific singleton instance in the dict and returns it, or creates and adds
    the instance to the dict and then returns it.

    To create a getter function, we need three pieces of information:
    a key for the singletons dict, a module where the singleton's class can
    be imported, and the name of that class. The module is given relative to
    the ``karabogui`` package. A getter function will also get a custom
    creator function which uses the above information to import and instantiate
    the singleton. This uses ``importlib.import_module`` and ``getattr`` to
    resolve the class object.

    Finally, each singleton's class is expected to have an __init__ method
    which takes a single keyword argument ``parent`` which in our case will
    always be the global ``QApplication`` instance.
    """
    def creator():
        module = importlib.import_module(modulename, 'karabogui')
        klass = getattr(module, klassname)

        app = QApplication.instance()
        instance = klass(parent=app)
        __singletons[key] = instance
        return instance

    def getter():
        if key in __singletons:
            return __singletons[key]
        return creator()

    return getter


# This is where all the singletons are defined
get_db_conn = _build_getter('db_conn', '.singletons.db_connection',
                            'ProjectDatabaseConnection')
get_manager = _build_getter('manager', '.singletons.manager', 'Manager')
get_mediator = _build_getter('mediator', '.singletons.mediator', 'Mediator')
get_network = _build_getter('network', '.singletons.network', 'Network')
get_panel_wrangler = _build_getter('panel_wrangler',
                                   '.singletons.panel_wrangler',
                                   'PanelWrangler')
get_selection_tracker = _build_getter('selection_tracker',
                                      '.singletons.selection_tracker',
                                      'SelectionTracker')
get_project_model = _build_getter('project_model', '.singletons.project_model',
                                  'ProjectViewItemModel')
get_topology = _build_getter('topology', '.topology.api', 'SystemTopology')
get_config = _build_getter('configuration', '.singletons.configuration',
                           'Configuration')

# Hide our implementation detail
del _build_getter



#############################################################################
# Author: <alessandro.silenzi@xfel.eu>
# Created on November 23, 2022
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from pathlib import Path

from qtpy.QtCore import Qt
from qtpy.QtSvg import QSvgWidget

import karabogui.const as global_constants
from karabo.common.api import InstanceStatus
from karabogui import icons
from karabogui.binding.api import ProxyStatus
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.singletons.api import get_topology
from karabogui.widgets.hints import KaraboSceneWidget

ICON_PATH = Path(icons.__file__).parent

SVG_OFFLINE = str(ICON_PATH.joinpath("statusOffline.svg"))
SVG_OK = str(ICON_PATH.joinpath("statusOk.svg"))
SVG_ERROR = str(ICON_PATH.joinpath("statusError.svg"))
SVG_UNKNOWN = str(ICON_PATH.joinpath("statusUnknown.svg"))

SVG_STATUS = {
    InstanceStatus.OK: SVG_OK,
    InstanceStatus.UNKNOWN: SVG_UNKNOWN,
    InstanceStatus.ERROR: SVG_ERROR,
    InstanceStatus.NONE: SVG_OFFLINE
}


class InstanceStatusWidget(KaraboSceneWidget, QSvgWidget):
    """A widget to display a device status on the scene"""

    def __init__(self, model, parent=None):
        self.device_id = model.keys[0].split(".", 1)[0]
        super().__init__(model=model, parent=parent)
        self.status = ProxyStatus.OFFLINE
        self.setCursor(Qt.PointingHandCursor)
        self.proxy = get_topology().get_device(self.device_id, False)
        self.proxy.on_trait_change(self.proxy_online_change, "online")
        self.proxy.on_trait_change(self.proxy_status_change,
                                   "topology_node:status")
        self.proxy_online_change(self.proxy.online)
        self.setGeometry(model.x, model.y, model.width, model.height)

    def mousePressEvent(self, event):
        if global_constants.APPLICATION_MODE:
            broadcast_event(KaraboEvent.RaiseEditor, {"proxy": self.proxy})
        else:
            broadcast_event(KaraboEvent.ShowDevice,
                            {"deviceId": self.device_id})

    def getToolTip(self):
        return f"{self.device_id} - {self.status.name}"

    def proxy_online_change(self, online):
        if online and self.proxy.topology_node is not None:
            svg = SVG_STATUS[self.proxy.topology_node.status]
            self.status = self.proxy.topology_node.status
        else:
            svg = SVG_OFFLINE
            self.status = ProxyStatus.OFFLINE
        self.setToolTip(self.getToolTip())
        self.load(svg)

    def proxy_status_change(self, status):
        svg = SVG_STATUS[status]
        self.status = status
        self.setToolTip(self.getToolTip())
        self.load(svg)

    def set_geometry(self, rect):
        """Satisfy the informal widget interface."""
        self.model.trait_set(x=rect.x(), y=rect.y(),
                             width=rect.width(), height=rect.height())
        self.setGeometry(rect)

    def translate(self, offset):
        """Satisfy the informal widget interface."""
        new_pos = self.pos() + offset
        self.model.trait_set(x=new_pos.x(), y=new_pos.y())
        self.move(new_pos)

    def add_proxies(self, proxies):
        """Satisfy the informal widget interface."""

    def apply_changes(self):
        """Satisfy the informal widget interface."""

    def decline_changes(self):
        """Satisfy the informal widget interface."""

    def destroy(self):
        """Satisfy the informal widget interface."""
        self.proxy.on_trait_change(self.proxy_status_change,
                                   "topology_node:status", remove=True)
        self.proxy.on_trait_change(self.proxy_online_change, "online",
                                   remove=True)
        super().destroy()

    def set_visible(self, visible):
        """Satisfy the informal widget interface."""

    def update_global_access_level(self, level):
        """Satisfy the informal widget interface."""



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from karabo.native import Configurable, String
from karabogui.testing import get_class_property_proxy, set_proxy_value

from ..lineedit import DisplayLineEdit


class Object(Configurable):
    prop = String()


def test_set_string_value(gui_app):
    # set up
    schema = Object.getClassSchema()
    proxy = get_class_property_proxy(schema, "prop")
    controller = DisplayLineEdit(proxy=proxy)
    controller.create(None)
    assert controller.widget is not None

    # body
    set_proxy_value(proxy, "prop", "hello")
    assert controller.widget.text() == "hello"

    # teardown
    controller.destroy()
    assert controller.widget is None



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import re
import sys
import urllib.request
from contextlib import closing
from pathlib import Path
from urllib.error import URLError

from qtpy import uic
from qtpy.QtCore import QBuffer, QByteArray, Qt, Signal, Slot
from qtpy.QtGui import QPalette, QPixmap, QPixmapCache
from qtpy.QtWidgets import (
    QApplication, QDialog, QLabel, QLineEdit, QSizePolicy)
from traits.api import Instance, Property

from karabo.common.scenemodel.api import IconData
from karabogui import icons, messagebox
from karabogui.binding.api import FloatBinding, get_min_max
from karabogui.singletons.api import get_config
from karabogui.util import getOpenFileName, temp_file
from karabogui.validators import IntValidator, NumberValidator

ICON_FILE_SIZE_LIMIT = 102400  # 100 KB
MIN_SIZE = 24
DEFAULT_SIZE = 100


class IconError(Exception):
    pass


class IconFileSizeLimitError(Exception):
    """Exception for icons exceeding file size limit"""


class IconLabel(QLabel):

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setMinimumSize(MIN_SIZE, MIN_SIZE)
        self.setSizePolicy(QSizePolicy.MinimumExpanding,
                           QSizePolicy.MinimumExpanding)
        self._pixmap = self._default_pixmap = icons.no.pixmap(100)

    # Qt Overrides
    # ---------------------------------------------------------------------

    def resizeEvent(self, event):
        """Reimplemented method to scale the QPixmap"""
        super().resizeEvent(event)
        self.setPixmap(self._pixmap)

    def setPixmap(self, pixmap):
        """Reimplemented method to store the input pixmap and use the
           scaled version"""
        self._pixmap = pixmap
        size = self.size()
        pixmap = pixmap.scaled(size, Qt.KeepAspectRatio,
                               Qt.SmoothTransformation)
        super().setPixmap(pixmap)

    def setDefaultPixmap(self):
        self.setPixmap(self._default_pixmap)


class Label(QLabel):
    """A Custom QLabel subclass which is referenced by the 'icons.ui' file

    This Label shrinks a Pixmap if necessary to the size, but by default
    does not enlarge a label if there is more space.
    """
    newMime = Signal(str)

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setAcceptDrops(True)
        self._pixmap = self._default_pixmap = icons.no.pixmap(100)

    # Qt Overrides
    # ---------------------------------------------------------------------

    def resizeEvent(self, event):
        """Reimplemented method to reset the pixmap scale with the
           widget resize"""
        if self._pixmap is not None:
            super().setPixmap(self._scaled_pixmap)

        super().resizeEvent(event)

    def setPixmap(self, pixmap):
        """Reimplemented method to store the input pixmap and use the
           scaled version internally"""
        self._pixmap = pixmap
        super().setPixmap(self._scaled_pixmap)

    def dragEnterEvent(self, event):
        event.acceptProposedAction()

    def dropEvent(self, event):
        self.newMime.emit(event.mimeData())

    # Properties
    # ---------------------------------------------------------------------

    @property
    def _scaled_pixmap(self):
        """Scales the pixmap only if the pixmap is larger than the widget"""
        pixmap = self._pixmap
        if self.needs_shrink:
            pixmap = pixmap.scaled(self.size(),
                                   Qt.KeepAspectRatio, Qt.SmoothTransformation)

        return pixmap

    @property
    def needs_shrink(self):
        """Checks if the pixmap is larger than the widget by comparing
           the widths and heights"""
        pixmap_size = self._pixmap.size()
        widget_size = self.size()
        horizontally_big = pixmap_size.width() > widget_size.width()
        vertically_big = pixmap_size.height() > widget_size.height()
        return horizontally_big or vertically_big

    def setDefaultPixmap(self):
        self.setPixmap(self._default_pixmap)


class IconItem(IconData):
    """Inherit from the scene model `IconData` class to give life to the traits
    which are already declared by that class.
    """
    pixmap = Property(Instance(QPixmap))
    re = Property()  # For TextIcons

    def _get_pixmap(self):
        # If no data is added, we use the default pixmap
        if not self.data:
            # If the default pixmap url is not yet existing, we add the pixmap
            pixmap = icons.no.pixmap(DEFAULT_SIZE, DEFAULT_SIZE)
            self._cache_pixmap(pixmap, update=True)
            return pixmap

        # If image does not have a temp url, the item is initially loaded from
        # the model. It needs to be saved in a temp file and has to be added
        # in the pixmap cache
        if not self.image:
            pixmap = self._load_pixmap()
            self._cache_pixmap(pixmap)
            return pixmap

        # Get pixmap from cache
        pixmap = QPixmapCache.find(self.image)

        # If the pixmap is not in the cache, we need to load and add the data.
        if pixmap is None:
            pixmap = self._load_pixmap()
            self._cache_pixmap(pixmap)

        return pixmap

    def _get_re(self):
        return re.compile(self.value)

    def _image_changed(self, old_url, new_url):
        """The `image` trait is the URL of the image data of this item"""
        if not new_url.startswith("file:"):
            new_url = "file://" + urllib.request.pathname2url(new_url)
        try:
            data = urllib.request.urlopen(new_url).read()

            # Check if data size is larger than the limit
            file_size = sys.getsizeof(data)
            if file_size > ICON_FILE_SIZE_LIMIT:
                raise IconFileSizeLimitError

            # Try to load the file as QPixmap to check image validity
            pixmap = QPixmap()
            if not pixmap.loadFromData(data):
                raise IconError
        except URLError:
            # Revert changes
            self.trait_setq(image=old_url)
            messagebox.show_error("Pasted image is invalid.")
        except IconFileSizeLimitError:
            self.trait_setq(image=old_url)
            message = ("Desired file exceeds the file size limit of "
                       "{size_limit} KB [{file_size:.2f} KB]."
                       .format(size_limit=int(ICON_FILE_SIZE_LIMIT / 1024),
                               file_size=file_size / 1024))
            messagebox.show_error(message)
        except IconError:
            self.trait_setq(image=old_url)
            messagebox.show_error("Could not read image.")
        else:
            self.data = data

    def _load_pixmap(self):
        """Load the pixmap from the saved bytes"""
        pixmap = QPixmap()
        try:
            if not pixmap.loadFromData(self.data):
                raise IconError
        except (KeyError, IconError):
            messagebox.show_error("Could not read image.")
            pixmap = icons.no.pixmap(DEFAULT_SIZE, DEFAULT_SIZE)

        return pixmap

    def _cache_pixmap(self, pixmap, update=False):
        """Save the pixmap in the cache, with a temporary file as key.
           Optionally update the data, which is usually not desired as it is
           already previously loaded."""
        temp_url = self.create_temp_url(pixmap, update=update)
        QPixmapCache.insert(temp_url, pixmap)

    def create_temp_url(self, image_data, update=True):
        """Create temporary URL from given ``image_data``."""
        ba = QByteArray()
        buff = QBuffer(ba)
        buff.open(QBuffer.WriteOnly)
        with closing(buff):
            image_data.save(buff, "PNG")
            data = buff.data()
            with temp_file() as tmp_path:
                with open(tmp_path, "wb") as out:
                    out.write(data)

        if update:
            self.data = data.data()
        self.trait_setq(image=tmp_path)

        return tmp_path


class _BaseIconDialog(QDialog):
    def __init__(self, items, parent=None):
        super().__init__(parent)
        uic.loadUi(Path(__file__).parent.joinpath("icons.ui"), self)
        # Get a copy!
        self.items = list(items)
        self.valueList.addItems([self.text_for_item(item) for item in items])
        self.valueList.setCurrentRow(0)
        self._update_buttons()

    def _update_buttons(self):
        """Only allow Open and Paste buttons if the rows contains something
        """
        enabled = len(self.items) > 0
        self.open.setEnabled(enabled)
        self.paste.setEnabled(enabled)
        self.deleteValue.setEnabled(enabled)

    def _update_image(self, item):
        """Update the ``self.image`` with the given pixmap of the ``item``.
        """
        pixmap = item.pixmap
        if pixmap is None:
            return
        self.image.setPixmap(pixmap)

    @Slot(int)
    def on_valueList_currentRowChanged(self, row):
        """Update the buttons on every selection change and select the pixmap
        """
        self._update_buttons()
        self.image.setPixmap(self.items[row].pixmap)

    @Slot()
    def on_open_clicked(self):
        """Slot executing on `open` button"""
        filename = getOpenFileName(
            parent=self, caption="Open Icon",
            filter="Images (*.png *.xpm *.jpg *.jpeg *.svg *.gif *.ico "
                   "*.tif *.tiff *.bmp)", directory=get_config()["data_dir"])
        if not filename:
            return

        get_config()["data_dir"] = filename
        item = self.items[self.valueList.currentRow()]
        item.image = filename
        self._update_image(item)

    @Slot()
    def on_paste_clicked(self):
        """Slot executing on `paste` button"""
        mime = QApplication.clipboard().mimeData()
        item = self.items[self.valueList.currentRow()]
        if mime.hasImage():
            item.create_temp_url(mime.imageData())
        else:
            try:
                filename = mime.text().strip()
            except Exception as e:
                e.message = "Could not open URL or Image"
                raise
            item.image = filename
        self._update_image(item)

    @Slot()
    def on_deleteValue_clicked(self):
        """Slot executing on `delete` button"""
        row = self.valueList.currentRow()
        if self.items[row].value is not None:
            self.valueList.takeItem(row)
            del self.items[row]
            if len(self.items) == 0:
                self.image.setDefaultPixmap()
        self._update_buttons()

    # --------------------------------------------------------------------
    # Abstract interface

    def text_for_item(self, item):
        pass


class LineEditEditor(QLineEdit):

    def __init__(self, parent=None):
        super().__init__(parent)
        self._normal_palette = self.palette()
        self._error_palette = QPalette(self._normal_palette)
        self._error_palette.setColor(QPalette.Text, Qt.red)
        self.textChanged.connect(self._check_background)

    @Slot(str)
    def _check_background(self, text):
        acceptable_input = self.hasAcceptableInput()
        palette = (self._normal_palette if acceptable_input
                   else self._error_palette)
        self.setPalette(palette)


class DigitDialog(_BaseIconDialog):
    def __init__(self, items, binding, parent=None):
        super().__init__(items, parent)
        self.value = LineEditEditor(parent=self)
        minInc, maxInc = get_min_max(binding)
        if isinstance(binding, FloatBinding):
            validator = NumberValidator(minInc, maxInc)
            self.float_layout.insertWidget(0, self.value)
            self.stack.setCurrentWidget(self.floatPage)
            self.cast = float
        else:
            validator = IntValidator(minInc, maxInc)
            self.int_layout.insertWidget(0, self.value)
            self.stack.setCurrentWidget(self.intPage)
            self.cast = int
        # Set the validator on the line edit
        self.value.setValidator(validator)

    @Slot()
    def on_addValue_clicked(self):
        if not self.value.hasAcceptableInput():
            text = (f"Value {self.value.text()} is not a valid value or in "
                    "the range of the property limits.")
            messagebox.show_error(text, parent=self)
            return

        new_idx = 0
        entry_value = self.cast(self.value.text())
        entry_equal = self.lessEqual.isChecked()
        for idx, item in enumerate(self.items):
            # Check if entry is already existing
            item_value = float(item.value)
            if entry_value == item_value and entry_equal is item.equal:
                message = "Cannot add new condition; it already exists."
                messagebox.show_error(message, parent=self)
                return

            # Find new position in list widget
            if (entry_value < item_value or
                    entry_value == item_value and item.equal):
                new_idx = idx
                break

            new_idx = idx + 1

        item = IconItem(value=entry_value, equal=entry_equal)
        self.items.insert(new_idx, item)
        self.valueList.insertItem(new_idx, self.text_for_item(item))
        # Trigger the imageView to generate default data!
        self.valueList.setCurrentRow(new_idx)

    def text_for_item(self, item):
        if item.value is None:
            return "default"
        return "{} {}".format("<=" if item.equal else "<", item.value)


class SelectionDialog(_BaseIconDialog):
    def __init__(self, items, binding, parent=None):
        super().__init__(items, parent)
        self.editable.hide()

    def text_for_item(self, item):
        return item.value


class TextDialog(_BaseIconDialog):
    def __init__(self, items, binding, parent=None):
        super().__init__(items, parent)
        self.stack.setCurrentWidget(self.textPage)

        if binding.options:
            self.stack.setCurrentWidget(self.textOptionsPage)
            self.itemsComboBox.addItems(list(binding.options))

    @Slot()
    def on_addValue_clicked(self):
        if self.stack.currentWidget() == self.textPage:
            entry_value = self.textValue.text().strip()
        else:
            entry_value = self.itemsComboBox.currentText()
        # Check if entry is already existing
        if entry_value in [item.value for item in self.items]:
            message = "Cannot add new condition; it already exists."
            messagebox.show_error(message, parent=self)
            return

        item = IconItem(value=entry_value)
        self.items.insert(0, item)
        self.valueList.insertItem(0, entry_value)
        # Trigger the imageView to generate default data!
        self.valueList.setCurrentRow(0)

    def text_for_item(self, item):
        if not item.value:
            return "default"
        return item.value



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import os.path as op

from qtpy.QtGui import QIcon


class Icon:
    """A lazy-loading class for QIcons

    QIcons can only be read once a QApplication is created. This
    class assures that icons are only loaded when needed."""
    def __init__(self, name):
        self.name = name
        self._icon = None

    @property
    def icon(self):
        if self._icon is None:
            self._icon = QIcon(op.join(op.dirname(__file__), self.name))
        return self._icon

    def __get__(self, instance, owner):
        return self.icon


def init():
    """This function actually loads the hitherto lazily loaded icons."""
    d = globals()
    for k, v in d.items():
        if isinstance(v, Icon):
            d[k] = v.icon

    global applyGrey
    applyGrey = QIcon(apply.pixmap(32, QIcon.Disabled, QIcon.On))


about = Icon("about.svg")
add = Icon("add.svg")
alarmGlobal = Icon("critical.svg")
alarmHigh = Icon("critical.svg")
alarmLow = Icon("critical.svg")
alarmVarianceHigh = Icon("critical.svg")
alarmVarianceLow = Icon("critical.svg")
alarmWarning = Icon("alarmwarning.svg")
alignBottom = Icon("align-bottom.svg")
alignLeft = Icon("align-left.svg")
alignRight = Icon("align-right.svg")
alignTop = Icon("align-top.svg")
apply = Icon("apply-32x32.png")
applyConflict = Icon("apply-conflict-32x32.png")
arrow = Icon("arrow.svg")
arrowDown = Icon("arrow-down.svg")
arrowLeft = Icon("arrow-left.svg")
arrowRight = Icon("arrow-right.svg")
arrowUp = Icon("arrow-up.svg")
arrowFancyDown = Icon("fancy-arrow-down.svg")
arrowFancyLeft = Icon("fancy-arrow-left.svg")
arrowFancyRight = Icon("fancy-arrow-right.svg")
arrowFancyUp = Icon("fancy-arrow-up.svg")
beamProfile = Icon("beam-profile.svg")
boolean = Icon("type-boolean.png")
bound = Icon("bound-logo.svg")
boundDevelopment = Icon("bound-logo-development.svg")
bringToFront = Icon("bring-to-front1-32x32.png")
camera = Icon("camera.svg")
change = Icon("change.svg")
consoleMenu = Icon("console-menu.svg")
clock = Icon("clock.svg")
close = Icon("close-icon.svg")
close_small = Icon("close-icon-20x20.svg")
crosshair = Icon("crosshair.svg")
cursorArrow = Icon("cursor-arrow-32x32.png")
cpp = Icon("cpp-logo.svg")
cppDevelopment = Icon("cpp-logo-development.svg")
crop = Icon("crop.svg")
data_analysis = Icon("data_analysis.svg")
delete = Icon("edit-remove.svg")
deviceAlive = Icon("device-alive.svg")
deviceClass = Icon("device-class.png")
deviceIncompatible = Icon("device-incompatible.png")
deviceInstance = Icon("device-instance.png")
deviceInstanceError = Icon("device-instance-error.png")
deviceMonitored = Icon("device-monitored.svg")
deviceOffline = Icon("offline-32x32.png")
deviceOfflineNoPlugin = Icon("offline-no-plugin.png")
deviceOfflineNoServer = Icon("offline-no-server.png")
device_dead = Icon("device-dead.svg")
device_error = Icon("device-error.svg")
device_requested = Icon("device-requested.svg")
device_schema = Icon("device-schema.svg")
dock = Icon("down-32x32.png")
download = Icon("download.svg")
downsample = Icon("downsample.svg")
draw = Icon("draw.svg")
drawROI = Icon("draw-roi.svg")
drawCrosshair = Icon("draw-crosshair.svg")
edit = Icon("edit.svg")
editClear = Icon("edit-clear-32x32.png")
editCopy = Icon("edit-copy-32x32.png")
editCut = Icon("edit-cut-32x32.png")
editPaste = Icon("edit-paste.svg")
editPasteReplace = Icon("edit-paste-replace.svg")
entireWindow = Icon("entire.svg")
enum = Icon("type-list.png")
eraser = Icon("eraser.svg")
exit = Icon("exit-32x32.png")
export = Icon("export.svg")
file = Icon("file.svg")
filein = Icon("filein-32x32.png")
fileout = Icon("fileout-32x32.png")
float = Icon("type-float.png")
folder = Icon("folder.svg")
folderDomain = Icon("folder-domain")
folderTrash = Icon("folder-trash.svg")
folderType = Icon("folder-type.svg")
grafana = Icon("grafana.svg")
grid = Icon("grid.svg")
group = Icon("group-32x32.png")
groupGrid = Icon("group-grid.svg")
groupHorizontal = Icon("group-horizontal.svg")
groupVertical = Icon("group-vertical.svg")
halt = Icon("halt.svg")
histogram = Icon("histogram.svg")
host = Icon("computer-32x32.png")
image = Icon("image-gray-32x32.png")
imageMissing = Icon("image-missing.svg")
info = Icon("info.svg")
initconfig = Icon("init-config.svg")
inspect_code = Icon("inspect.svg")
inspect_code_error = Icon("inspect-error.svg")
inspect_code_ok = Icon("inspect-ok.svg")
int = Icon("type-integer.png")
interlock = Icon("interlock.svg")
invisibleEye = Icon("invisibleEye.svg")
kill = Icon("delete-32x32.png")
line = Icon("line.svg")
link = Icon("link-32x32.png")
load = Icon("file-open.svg")
lock = Icon("locked.svg")
logo = Icon("splash.png")
logbook = Icon("electronic-log.svg")
logAlarm = Icon("log-alarm-32x32.png")
logDebug = Icon("log-debug-32x32.png")
logError = Icon("log-error-32x32.png")
logInfo = Icon("log-info-32x32.png")
logMenu = Icon("log-menu.svg")
logWarning = Icon("log-warning-32x32.png")
macro = Icon("macro-logo.svg")
macroDevelopment = Icon("macro-logo-development.svg")
homeEdit = Icon("homeedit.svg")
maximize = Icon("view-maximize.svg")
mediaBackward = Icon("media-reset.svg")
mediaPause = Icon("media-pause.svg")
mediaRecord = Icon("media-record.svg")
mediaStart = Icon("media-start.svg")
mediaStop = Icon("media-stop.svg")
minimize = Icon("view-minimize.svg")
move = Icon("move.svg")
new = Icon("filenew-32x32.png")
no = Icon("no.png")
note_info = Icon("info.svg")
note_warning = Icon("info-help.svg")
note_alert = Icon("info-critical.svg")
off = Icon("off.svg")
on = Icon("on.svg")
path = Icon("type-string.png")
picker = Icon("picker.svg")
plus = Icon("plus.svg")
pointer = Icon("pointer.svg")
popup_sticker = Icon("popup-sticker.svg")
printer = Icon("printer.svg")
propertyMissing = Icon("property-missing.svg")
python = Icon("python-logo.svg")
pythonDevelopment = Icon("python-logo-development.svg")
rect = Icon("rect-32x32.png")
refresh = Icon("refresh-32x32.png")
reset = Icon("reset.svg")
resetTilted = Icon("reset-tilted.svg")
remote = Icon("remote.png")
removeUser = Icon("remove_user.svg")
minus = Icon("minus.svg")
resize = Icon("resize.svg")
revert = Icon("edit-undo.png")
roi = Icon("roi.svg")
run = Icon("run.png")
save = Icon("filesave-32x32.png")
saveAs = Icon("filesaveas-32x32.png")
scatter = Icon("scatter.svg")
scenelink = Icon("scenelink-32x32.png")
selectAll = Icon("select-all.svg")
sendToBack = Icon("bring-to-back1-32x32.png")
settings = Icon("settings.svg")
signal = Icon("type-schema.png")
slot = Icon("type-pair.png")
split = Icon("split.svg")
statusOk = Icon("statusOk.svg")
statusError = Icon("statusError.svg")
statusUnknown = Icon("statusUnknown.svg")
sticker = Icon("sticker.svg")
stop = Icon("stop.svg")
string = Icon("type-string.png")
stringAttribute = Icon("type-string-attribute.png")
switchNormal = Icon("lock_switch.svg")
switchActive = Icon("lock_switch_active.svg")
switchWarning = Icon("lock_switch_warning.svg")
switchCritical = Icon("lock_switch_critical.svg")
target = Icon("target.svg")
text = Icon("text-32x32.png")
transform = Icon("transform-32x32.png")
undefined = Icon("type-undefined.png")
undefinedAttribute = Icon("type-undefined-attribute.png")
undock = Icon("up-32x32.png")
ungroup = Icon("ungroup-32x32.png")
warnGlobal = Icon("warning.svg")
warnHigh = Icon("warning.svg")
warnLow = Icon("warning.svg")
warnVarianceHigh = Icon("warning.svg")
warnVarianceLow = Icon("warning.svg")
weblink = Icon("weblink.svg")
yes = Icon("yes.png")
zoom = Icon("zoom.svg")
zoomImage = Icon("zoom-image.svg")
zoomIn = Icon("zoom-in.svg")
zoomOut = Icon("zoom-out.svg")



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pathlib import Path

from qtpy import uic
from qtpy.QtCore import Qt, Signal, Slot
from qtpy.QtGui import QClipboard, QIntValidator, QPixmap
from qtpy.QtWidgets import (
    QApplication, QGroupBox, QHBoxLayout, QLabel, QLineEdit, QListWidget,
    QListWidgetItem, QPushButton, QRadioButton, QSizePolicy, QSpacerItem,
    QVBoxLayout, QWizard, QWizardPage)

from karabogui.util import SignalBlocker

from .const import LOGO_PATH, LOGO_WIDTH

LABEL_HEIGHT = 20


class PageContainer(QWizard):
    def __init__(self, parent=None):
        super().__init__(parent=parent)
        self.setAttribute(Qt.WA_DeleteOnClose)
        pixmap = QPixmap(LOGO_PATH)
        pixmap = pixmap.scaledToWidth(LOGO_WIDTH)
        self.setPixmap(QWizard.LogoPixmap, pixmap)
        self.currentIdChanged[int].connect(self.fitContents)

    def setPage(self, index, page):
        page.container = self
        super().setPage(index, page)

    @Slot(int)
    def fitContents(self, index):
        self.adjustSize()


class Page(QWizardPage):
    header = None
    subheader = None
    container = None
    ui_file = None

    def __init__(self, parent=None):
        super().__init__(parent=parent)
        if self.ui_file is not None:
            ui_path = Path(__file__).parent / self.ui_file
            uic.loadUi(str(ui_path), self)
        self.setAttribute(Qt.WA_DeleteOnClose)

        # Initialize variables
        self._is_complete = True
        if self.header is not None:
            self.setTitle(self.header)
        if self.subheader is not None:
            self.setSubTitle(self.subheader)

        self.setup_page()

    def set_complete(self, complete):
        if self._is_complete != complete:
            self._is_complete = complete
            self.completeChanged.emit()

    def set_header(self, text):
        self.header = text
        self.setTitle(text)

    def set_subheader(self, text):
        self.subheader = text
        self.setSubTitle(text)

    def isComplete(self):
        return self._is_complete

    # Abstract interface
    # -----------------------------------------------------------------------

    def setup_page(self):
        """Reimplement to populate the page with the desired contents"""


LEFT_ARROW = "\u2190"
RIGHT_ARROW = "\u2192"


class SelectScenesPage(Page):

    def sceneAdded(self, model):
        """An empty handler to be configured"""

    def sceneRemoved(self, model):
        """An empty handler to be configured"""

    def setup_page(self):
        # Available scenes
        available_label = QLabel("Available Scenes:")
        self.available_listwidget = avail_widget = QListWidget()
        avail_widget.itemSelectionChanged.connect(self._available_selected)
        self.add_button = QPushButton(f"Add {RIGHT_ARROW}")
        self.add_button.setEnabled(False)
        self.add_button.clicked.connect(self._add_scene)
        spacer = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Fixed)
        add_hbox = QHBoxLayout()
        add_hbox.addWidget(self.add_button)
        add_hbox.addSpacerItem(spacer)
        available_vbox = QVBoxLayout()
        available_vbox.addWidget(available_label)
        available_vbox.addWidget(avail_widget)
        available_vbox.addLayout(add_hbox)

        # Selected scenes
        selected_label = QLabel("Selected Scenes:")
        self.selected_listwidget = select_widget = QListWidget()
        select_widget.itemSelectionChanged.connect(self._selected_selected)
        self.remove_button = QPushButton(f"{LEFT_ARROW} Remove")
        self.remove_button.setEnabled(False)
        self.remove_button.clicked.connect(self._remove_scene)
        spacer = QSpacerItem(0, 0, QSizePolicy.Expanding, QSizePolicy.Fixed)

        remove_hbox = QHBoxLayout()
        remove_hbox.addWidget(self.remove_button)
        remove_hbox.addSpacerItem(spacer)
        selected_vbox = QVBoxLayout()
        selected_vbox.addWidget(selected_label)
        selected_vbox.addWidget(select_widget)
        selected_vbox.addLayout(remove_hbox)

        # Combine everything
        content_hbox = QHBoxLayout(self)
        content_hbox.addLayout(available_vbox)
        content_hbox.addLayout(selected_vbox)

        self.setLayout(content_hbox)

    def set_available_scenes(self, scenes):
        self.available_listwidget.clear()
        for text, data in scenes:
            item = QListWidgetItem(text)
            item.setData(Qt.UserRole, data)
            self.available_listwidget.addItem(item)

    def set_selected_scenes(self, scenes):
        self.selected_listwidget.clear()
        for text, data in scenes:
            item = QListWidgetItem(text)
            item.setData(Qt.UserRole, data)
            self.selected_listwidget.addItem(item)

    @Slot()
    def _available_selected(self):
        selected = self.available_listwidget.selectedItems()
        self.add_button.setEnabled(bool(len(selected)))
        # Update opposite list widget
        with SignalBlocker(self.selected_listwidget):
            self.selected_listwidget.clearSelection()
        self.remove_button.setEnabled(False)

    @Slot()
    def _selected_selected(self):
        selected = self.selected_listwidget.selectedItems()
        self.remove_button.setEnabled(bool(len(selected)))
        # Update opposite list widget
        with SignalBlocker(self.available_listwidget):
            self.available_listwidget.clearSelection()
        self.add_button.setEnabled(False)

    @Slot()
    def _add_scene(self):
        list_widget = self.available_listwidget
        selected = list_widget.selectedItems()
        if not len(selected):
            return

        # We are sure that only one text is selected
        selected = selected[0]
        list_widget.takeItem(list_widget.row(selected))
        self.selected_listwidget.addItem(selected)
        self.sceneAdded(selected.data(Qt.UserRole))

    @Slot()
    def _remove_scene(self):
        list_widget = self.selected_listwidget
        selected = list_widget.selectedItems()
        if not len(selected):
            return

        # We are sure that only one text is selected
        selected = selected[0]
        list_widget.takeItem(list_widget.row(selected))
        self.available_listwidget.addItem(selected)
        self.sceneRemoved(selected.data(Qt.UserRole))


class ConfigurePage(Page):
    header = "Create Cinema Link"
    subheader = "..for scenes in this project with the meta data"

    ui_file = "link.ui"

    configChanged = Signal(object)

    edits = None

    def setup_page(self):
        self.host_info_widget.setVisible(False)
        self.splash_checkbox.toggled.connect(self._show_splash_changed)
        self.include_host_checkbox.toggled.connect(self._show_host_info)
        self.host_lineedit.textChanged.connect(self._host_changed)
        self.port_lineedit.setValidator(QIntValidator())
        self.port_lineedit.textChanged.connect(self._port_changed)

        # Register fields
        self.edits = [self.host_lineedit, self.port_lineedit]
        for field in self.edits:
            field.textChanged.connect(self._check_empty_fields)

    @Slot()
    def _check_empty_fields(self):
        """validation of empty input fields"""
        complete = all([field.text() for field in self.edits])
        self.set_complete(complete)

    def set_config(self, config):
        self.splash_checkbox.setChecked(config["show_splash"])
        self.include_host_checkbox.setChecked(config["include_host"])
        self.host_lineedit.setText(config["host"])
        self.port_lineedit.setText(str(config["port"]))

    @Slot(bool)
    def _show_host_info(self, enabled):
        self.host_info_widget.setVisible(enabled)
        self.configChanged.emit({"include_host": enabled})

    @Slot(bool)
    def _show_splash_changed(self, enabled):
        self.configChanged.emit({"show_splash": enabled})

    @Slot()
    def _host_changed(self):
        self.configChanged.emit({"host": self.host_lineedit.text()})

    @Slot()
    def _port_changed(self):
        self.configChanged.emit({"port": self.port_lineedit.text()})


LINK_TEXT = [
    "karabo://cinema?domain=DOMAIN&scene_uuid= ...",
    "<a href='karabo://cinema?domain=DOMAIN ...'>Cinema Name</a>",
]


class LinkPage(Page):
    header = "Cinema Link Overview"
    subheader = "The following parameters have been configured ..."
    url = None

    def setup_page(self):
        vbox = QVBoxLayout(self)
        self.label = QLabel(self)
        self.label.setWordWrap(True)
        self.uuid_widget = QListWidget(self)
        description = QLabel("The link will be copied to the "
                             "clipboard on <b>Finish</b>.", self)
        vbox.addWidget(self.label)
        vbox.addWidget(self.uuid_widget)

        group = QGroupBox()
        self.classic = QRadioButton("Karabo URL", self)
        self.weblink = QRadioButton("HTML Link", self)
        self.classic.clicked.connect(self.update_url_info)
        self.weblink.clicked.connect(self.update_url_info)
        self.link_description = QLabel(self)
        self.link_description.setTextFormat(Qt.PlainText)

        name_layout = QHBoxLayout()
        self.name_label = QLabel("<b>Cinema Name</b>", self)
        self.cinema_name = QLineEdit("link text", self)

        name_layout.addWidget(self.name_label)
        name_layout.addWidget(self.cinema_name)

        # Set defaults
        self.link_description.setText(LINK_TEXT[0])
        self.classic.setChecked(True)
        self.cinema_name.setVisible(False)
        self.cinema_name.setMinimumHeight(25)
        self.name_label.setVisible(False)

        radio_layout = QHBoxLayout()
        radio_layout.addWidget(self.classic)
        radio_layout.addWidget(self.weblink)

        group_layout = QVBoxLayout()
        group_layout.addLayout(radio_layout)
        group_layout.addWidget(self.link_description)
        group_layout.addLayout(name_layout)

        group.setLayout(group_layout)
        group.adjustSize()
        vbox.addWidget(group)
        vbox.addWidget(description)

    @Slot(bool)
    def update_url_info(self, checked):
        if self.classic.isChecked():
            self.link_description.setText(LINK_TEXT[0])
            self.cinema_name.setVisible(False)
            self.name_label.setVisible(False)
        elif self.weblink.isChecked():
            self.link_description.setText(LINK_TEXT[1])
            self.cinema_name.setVisible(True)
            self.name_label.setVisible(True)

    def set_data(self, data):
        blacklist = (["uuids"] if data["include_host"]
                     else ["uuids", "host", "port"])
        meta = ("<table>" +
                "".join("<tr><td><b>{}</b>:   </td><td>{}</td></tr>".
                        format(key, str(value))
                        for key, value in data.items() if key not in blacklist)
                + "</table>")
        self.label.setText(meta)

        domain = data["domain"]
        include_host = data["include_host"]
        host = data["host"]
        port = data["port"]
        show_splash = data["show_splash"]
        uuids = data["uuids"]

        # Show the uuids in a list
        for data in uuids:
            item = QListWidgetItem(data)
            item.setData(Qt.UserRole, data)
            self.uuid_widget.addItem(item)

        # Mandatory args: domain and uuids
        args = ([f"domain={domain}"]
                + [f"scene_uuid={uuid}" for uuid in uuids])

        # Optional args: login details
        if include_host:
            args.extend([
                         f"host={host}",
                         f"port={port}",
                         ])
        # Optional args: no splash
        if not show_splash:
            args.append("nosplash")

        self.url = "karabo://cinema?" + "&".join(args)

    @Slot()
    def finished(self):
        if self.weblink.isChecked():
            self.url = f'<a href="{self.url}">{self.cinema_name.text()}</a>'
        clipboard = QApplication.clipboard()
        # Erase clipboard first!
        clipboard.clear(mode=QClipboard.Clipboard)
        clipboard.setText(self.url, mode=QClipboard.Clipboard)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
import argparse
import sys

from qtpy.QtCore import Slot

import karabogui.const as global_constants
from karabo.common.scenemodel.api import SceneTargetWindow
from karabogui.events import KaraboEvent, broadcast_event
from karabogui.messagebox import show_error
from karabogui.programs.base import create_gui_app, init_gui
from karabogui.programs.utils import close_app
from karabogui.singletons.api import get_db_conn, get_network, get_topology


def create_cinema(ns):
    app = create_gui_app(sys.argv)
    init_gui(app, use_splash=not ns.nosplash)

    def trigger_scenes():
        topology.system_tree.on_trait_change(
            trigger_scenes, 'initialized', remove=True)
        domain = ns.domain
        get_db_conn().default_domain = domain
        for uuid in ns.scene_uuid:
            db_scene = {'name': f'{domain}-{uuid}',
                        'target_window': SceneTargetWindow.MainWindow,
                        'target': uuid}
            broadcast_event(KaraboEvent.OpenSceneLink, db_scene)

    global_constants.APPLICATION_MODE = True

    topology = get_topology()
    # Attach to the topology
    topology.system_tree.on_trait_change(
        trigger_scenes, 'initialized')
    network = get_network()

    @Slot()
    def _connect_handler(status):
        """Connect handler for direct connect attempts without dialog"""
        network.signalServerConnectionChanged.disconnect(_connect_handler)
        if not status:
            show_error("Error, could not connect to gui server "
                       f"<b>{ns.host}:{ns.port}</b>. Closing karabo cinema.")
            topology.system_tree.on_trait_change(
                trigger_scenes, 'initialized', remove=True)
            close_app()

    # We might want to connect directly to the gui server
    if ns.host and ns.port:
        network.signalServerConnectionChanged.connect(_connect_handler)
        success = network.connectToServerDirectly(
           hostname=ns.host, port=ns.port)
    else:
        # Connect to the GUI Server via dialog
        success = network.connectToServer()

    return success, app


def run_cinema(ns):
    """The cinema is meant to directly download a scene from the project db

    From the initial scene the operator is allowed to request additional
    scene or device scene links.

    All scenes have the name ProjectDB|SceneName and are not editable!
    """
    success, app = create_cinema(ns)
    if success:
        app.exec()
        app.deleteLater()
        sys.exit()
    else:
        close_app()


def main():
    ap = argparse.ArgumentParser(description='Karabo Cinema')
    ap.add_argument('domain', type=str,
                    help='The domain where to look for the initial scene')
    ap.add_argument('scene_uuid', type=str, nargs='+',
                    help='The uuids of the scenes. This can be either a single'
                         'uuid or a sequence of uuids separated with a space')
    ap.add_argument('-host', '--host', type=str,
                    help='The hostname of the gui server to connect')
    ap.add_argument('-port', '--port', type=int,
                    help='The port number of the gui server to connect')
    ap.add_argument('-nosplash', '--nosplash', action='store_true')
    run_cinema(ap.parse_args())


if __name__ == '__main__':
    main()



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on November 24, 2016
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from collections import defaultdict

from qtpy.QtCore import QObject

from karabogui.events import KaraboBroadcastEvent


class Mediator(QObject):
    """The main dispatch point for KaraboBroadcastEvent events.

    See ``karabogui.events`` for more details.
    """

    def __init__(self, parent=None):
        super().__init__(parent)

        self._listeners = defaultdict(set)

    def event(self, event):
        """Dispatch KaraboBroadcastEvents to all interested listeners.
        """
        if not isinstance(event, KaraboBroadcastEvent):
            return super().event(event)

        sender = event.sender
        data = event.data
        # Make a copy in case listeners are added or removed!
        handlers = list(self._listeners.get(sender, ()))
        for handler in handlers:
            handler(data)

        event.accept()
        return True

    def register_listener(self, event_map):
        """Add an event map.
        """
        for event, handler in event_map.items():
            self._listeners[event].add(handler)

    def unregister_listener(self, event_map):
        """Remove an event map.
        """
        for event, handler in event_map.items():
            # Note: We should not ask to remove a handler that is there ...
            # However, in mocking tests and not graceful stops of the GUI we
            # protect here!
            self._listeners[event].discard(handler)
            if not self._listeners[event]:
                del self._listeners[event]



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
# flake8: noqa
from .schema import (
    ALL_PROPERTIES_MAP, get_all_props_schema, get_pipeline_schema,
    get_pipeline_vector_schema, get_simple_props_schema, get_simple_schema,
    get_slotted_schema, get_vectorattr_schema)
from .utils import (
    GuiTestCase, SimpleDeviceSchema, access_level, assert_no_throw,
    assert_trait_change, check_renderer_against_svg, click_button,
    development_system_hash, device_hash, flushed_registry,
    get_class_property_proxy, get_device_schema,
    get_device_schema_allowed_state, get_property_proxy, keySequence,
    lazy_scene_reading, set_proxy_hash, set_proxy_value,
    set_test_organization_info, singletons, system_hash,
    system_hash_server_and_plugins)



#############################################################################
# Author: <john.wiggins@xfel.eu>
# Created on June 1, 2017
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
from textwrap import dedent

from qtpy.QtCore import QEventLoop, QPoint, Qt
from qtpy.QtGui import QCursor
from qtpy.QtWidgets import QApplication, QMessageBox

from karabogui.dialogs.messagebox import KaraboMessageBox
from karabogui.singletons.api import get_panel_wrangler


def show_alarm(text, title="Alarm", details=None, parent=None):
    return _show_message_box(QMessageBox.Critical, text, title,
                             details, parent)


def show_error(text, title="Error", details=None, parent=None):
    return _show_message_box(QMessageBox.Critical, text, title,
                             details, parent)


def show_information(text, title="Information", details=None, parent=None):
    return _show_message_box(QMessageBox.Information, text, title,
                             details, parent)


def show_warning(text, title="Warning", details=None, parent=None):
    return _show_message_box(QMessageBox.Warning, text, title,
                             details, parent)


def _show_message_box(icon, text, title, details=None, parent=None):
    """A wrapper to simplify the different message box styles defined below.
    """
    message_box = KaraboMessageBox(parent=parent)
    message_box.setWindowTitle(title)
    message_box.setIcon(icon)
    message_box.setText(dedent(text))
    if details is not None:
        message_box.setDetailedText(details)

    if parent is not None:
        return message_box.open()

    _move_to_main_window(message_box)
    return message_box.exec()


def _move_to_main_window(msg_box):
    """Move a message box to the center of main window if available"""
    main_window = get_panel_wrangler().main_window
    if main_window is None or main_window.windowState() & Qt.WindowMinimized:
        return
    main_center = main_window.rect().center()
    msg_box_ref = QPoint(msg_box.width() // 2, msg_box.height() // 2)
    global_center = main_window.mapToGlobal(main_center - msg_box_ref)
    msg_box.move(global_center)


# --------------------------------------

def _move_to_cursor(msg_box):
    """Move the message box to the cursor"""
    pos = QCursor.pos()
    size = msg_box.size()

    x = pos.x() - size.width() // 2
    if x > 0:
        pos.setX(x)
    y = pos.y() - size.height() // 2
    if y > 0:
        pos.setY(y)
    msg_box.move(pos)


def _show_cursor_box(icon, text, title, details=None):
    message_box = KaraboMessageBox(parent=None)
    message_box.setWindowTitle(title)
    message_box.setIcon(icon)
    message_box.setText(dedent(text))
    if details is not None:
        message_box.setDetailedText(details)

    _move_to_cursor(message_box)
    QApplication.processEvents(QEventLoop.AllEvents, 100)
    return message_box.exec()


def show_error_at_cursor(text, title="Error", details=None):
    return _show_cursor_box(QMessageBox.Critical, text, title, details)



# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
from pyqtgraph import LabelItem, mkBrush, mkPen
from pyqtgraph.graphicsItems.LegendItem import ItemSample, LegendItem
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor, QPen

SYMBOL_COLUMN = 0
TEXT_COLUMN = 1

TEXT_SIZE = "8pt"
LEGEND_PEN = mkPen(QColor(192, 192, 192, 200))
LEGEND_BRUSH = mkBrush(QColor(0, 0, 0, 50))
LABEL_COLOR = QColor(0, 0, 0, 200)


class KaraboLegend(LegendItem):
    def __init__(self, size=None, offset=None, **kwargs):
        super().__init__(
            size, offset, pen=LEGEND_PEN, brush=LEGEND_BRUSH,
            labelTextColor=LABEL_COLOR)

    def addItem(self, item, name):
        """Reimplemented function of LegendItem

        :param item: A PlotDataItem from which the line and point style
                     of the item will be determined
        :param name: The title to display for this item. Simple HTML allowed.
        """
        label = LabelItem(name, justify='left',
                          color=self.opts['labelTextColor'],
                          size=TEXT_SIZE)
        sample = (item if isinstance(item, ItemSample)
                  else ColorBox(item))
        row = self.layout.rowCount()
        self.items.append((sample, label))
        self.layout.addItem(sample, row, SYMBOL_COLUMN)
        self.layout.addItem(label, row, TEXT_COLUMN)
        self.updateSize()


class CoordsLegend(KaraboLegend):
    def __init__(self, color='w'):
        super().__init__()
        self._label = LabelItem(color=color, size="8pt")
        self.layout.addItem(self._label, 0, 0)
        self.layout.setContentsMargins(2, 2, 2, 2)
        self.hide()

    def set_value(self, x, y):
        self._label.setText(f"x: {x}<br>y: {y}")


class ColorBox(ItemSample):
    """The color box in the legend that shows the curve pen color"""

    def __init__(self, item):
        super().__init__(item)
        self._pen = mkPen(item.opts.get('pen', None))
        self._brush = mkBrush(self._pen.color())

    def paint(self, painter, *args):
        painter.setPen(self._pen)
        painter.setBrush(self._brush)
        painter.drawRect(0, 0, 10, 14)

    def setPen(self, pen: QPen):
        self._pen = pen

    def setBrush(self, brush):
        self._brush = brush
        self.update()

    def mouseClickEvent(self, event):
        """Reimplemented function of PyQt

        Use the mouse click with the left button to toggle the visibility
        of the plot curve item as well as to change the appearance of the box
        """
        if event.button() == Qt.LeftButton:
            visible = self.item.isVisible()
            self.item.setVisible(not visible)
            color = QColor(211, 211, 211) if visible else self._pen.color()
            brush = mkBrush(color)
            self.setBrush(brush)

        event.accept()



#############################################################################
# Author: <dennis.goeries@xfel.eu>
# Created on June 14, 2019
# This file is part of the Karabo Gui.
#
# http://www.karabo.eu
#
# Copyright (C) European XFEL GmbH Schenefeld. All rights reserved.
#
# The Karabo Gui is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License, version 3 or higher.
#
# You should have received a copy of the General Public License, version 3,
# along with the Karabo Gui.
# If not, see <https://www.gnu.org/licenses/gpl-3.0>.
#
# The Karabo Gui is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.
#############################################################################
import os.path as op

from qtpy.QtWidgets import QAction
from traits.api import Instance, on_trait_change

from karabo.common.scenemodel.api import ErrorBoolModel
from karabogui import icons
from karabogui.binding.api import BoolBinding, get_binding_value
from karabogui.controllers.api import (
    BaseBindingController, register_binding_controller)
from karabogui.widgets.hints import SvgWidget

ICONS = op.dirname(icons.__file__)
OK_BOOL = op.join(ICONS, "ok-bool.svg")
ERROR_BOOL = op.join(ICONS, "error-bool.svg")


@register_binding_controller(ui_name='Ok-Error Bool',
                             klassname='DisplayErrorBool',
                             binding_type=BoolBinding, priority=0)
class DisplayErrorBool(BaseBindingController):
    # The scene data model class for this controller
    model = Instance(ErrorBoolModel, args=())

    def create_widget(self, parent):
        widget = SvgWidget(parent)
        widget.setMinimumSize(24, 24)

        logicAction = QAction("Invert color logic", widget)
        logicAction.triggered.connect(self.logic_action)
        widget.addAction(logicAction)
        logicAction.setCheckable(True)
        logicAction.setChecked(self.model.invert)

        return widget

    def value_update(self, proxy):
        value = get_binding_value(proxy, False)
        tooltip = f"{proxy.key}\n"

        if not self.model.invert:
            svg_file = OK_BOOL if value else ERROR_BOOL
        else:
            svg_file = ERROR_BOOL if value else OK_BOOL
            tooltip = f"{tooltip}Inverted: "

        tooltip = f"{tooltip}{value}"
        self.widget.setToolTip(tooltip)
        self.widget.load(svg_file)

    @on_trait_change('model.invert', post_init=True)
    def _invert_update(self):
        if self.proxy is not None:
            self.value_update(self.proxy)

    def logic_action(self):
        self.model.invert = not self.model.invert
